from basyx.aas.model import ConceptDescription


class ExtendedConceptDescription(ConceptDescription):
    """This class contains methods to be added to ConceptDescription class of Basyx Python SDK model."""

    def print_concept_description_information(self):
        print("Concept Description information:")
        print("\tid: {}".format(self.id))
        print("\tid_short: {}".format(self.id_short))
        print("\tdisplayName: {}".format(self.display_name))
        print("\tdescription: {}".format(self.description))
        print("\tcategory: {}".format(self.category))
        for embedded_ds in self.embedded_data_specifications:
            print("\tdataSpecification ref: {}".format(embedded_ds.data_specification))
            print("\tdataSpecification content:")
            print("\t\tpreferredName: {}".format(embedded_ds.data_specification_content.preferred_name))
            print("\t\tdefinition: {}".format(embedded_ds.data_specification_content.definition))




import logging

import aiohttp
from aiohttp import ClientConnectorError, ClientConnectionError
from basyx.aas.util import traversal

from smia.assetconnection.asset_connection import AssetConnection
from smia.logic.exceptions import AssetConnectionError
from smia.utilities.smia_info import AssetInterfacesInfo

_logger = logging.getLogger(__name__)


class HTTPAssetConnection(AssetConnection):
    """
    This class implements the asset connection for HTTP protocol.
    """

    def __init__(self):
        super().__init__()
        self.architecture_style = AssetConnection.ArchitectureStyle.CLIENTSERVER

        # Other data
        self.interface_title = None
        self.base = None
        self.endpoint_metadata_elem = None
        self.security_scheme_elem = None

        # Data of each request
        self.request_uri = None
        self.request_headers = {}
        self.request_params = None
        self.request_method = None
        self.request_body = None

    async def configure_connection_by_aas_model(self, interface_aas_elem):

        # The Interface element need to be checked
        await self.check_interface_element(interface_aas_elem)

        # Vamos a conseguir los datos necesarios del modelo AAS para configurar la conexion HTTP
        self.interface_title = interface_aas_elem.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACE_TITLE)
        # La informacion general de la conexion con el activo se define en el SMC 'EndpointMetadata'
        self.endpoint_metadata_elem = interface_aas_elem.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_ENDPOINT_METADATA)

        # The endpointMetadata element need to be checked
        await self.check_endpoint_metadata()

        self.base = self.endpoint_metadata_elem.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACE_BASE)
        content_type_elem = self.endpoint_metadata_elem.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACE_CONTENT_TYPE)
        if content_type_elem is not None:
            self.request_headers['Content-Type'] = content_type_elem.value

        security_definitions_elem = self.endpoint_metadata_elem.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACE_SECURITY_DEFINITIONS)
        if security_definitions_elem is not None:
            self.security_scheme_elem = security_definitions_elem.value
        # TODO: pensar como añadir el resto , p.e. tema de seguridad o autentificacion (bearer).
        #  De momento se ha dejado sin seguridad (nosec_sc)

        # The InteractionMetadata elements also need to be checked
        interaction_metadata_elem = interface_aas_elem.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERACTION_METADATA)
        for interaction_metadata_type in interaction_metadata_elem:
            # Interaction metadata can be properties, actions or events
            for interaction_element in interaction_metadata_type:
                await self.check_interaction_metadata(interaction_element)

    async def check_asset_connection(self):
        pass

    async def connect_with_asset(self):
        pass

    async def execute_asset_service(self, interaction_metadata, service_input_data = None):
        if interaction_metadata is None:
            raise AssetConnectionError("The skill cannot be executed by asset service because the given "
                                       "InteractionMetadata object is None", "invalid method parameter",
                                       "InteractionMetadata object is None")

        await self.extract_general_interaction_metadata(interaction_metadata)

        # Then, the data of the skill is added in the required field. To do that, the 'SkillParameterExposedThrough'
        # relationship should be obtained, which indicates where the parameter data should be added
        if service_input_data is not None:
            await self.add_asset_service_data(interaction_metadata, service_input_data)

        # At this point, the HTTP request is performed
        http_response = await self.send_http_request()
        if http_response:
            if http_response.status != 200:
                _logger.warning("The HTTP request has not been answered correctly. "
                                "Response: {}".format(await http_response.text()))
            return await self.get_response_content(interaction_metadata, await http_response.text())
        return None

    async def receive_msg_from_asset(self):
        pass

    # ---------------------
    # HTTP specific methods
    # ---------------------
    async def extract_general_interaction_metadata(self, interaction_metadata):
        """
        This method extracts the general interaction information from the interaction metadata object. Since this is an
        HTTP Asset Connection, information about the URI, headers and method name is obtained. All information is saved
        in the global variables of the class.

        Args:
             interaction_metadata (basyx.aas.model.SubmodelElementCollection): SubmodelElement of interactionMetadata.
        """
        # The interaction_metada element will be an SMC of the HTTP interface.
        await self.check_interaction_metadata(interaction_metadata)

        # First, the full URI of the HTTP request is obtained.
        forms_elem = interaction_metadata.get_sm_element_by_semantic_id(AssetInterfacesInfo.SEMANTICID_INTERFACE_FORMS)
        await self.get_complete_request_uri(forms_elem)

        # Then, headers are obtained
        await self.get_headers(forms_elem)

        # Eventually, the method name is also obtained
        await self.get_method_name(forms_elem)

    async def get_complete_request_uri(self, forms_elem):
        """
        This method gets the complete request URI from the forms element within the InteractionMetadata element. The
        information is saved in the global variables of the class.

        Args:
            forms_elem (basyx.aas.model.submodelElementCollection): SubmodelElement of forms within InteractionMetadata.
        """
        href_elem = forms_elem.get_sm_element_by_semantic_id(AssetInterfacesInfo.SEMANTICID_INTERFACE_HREF)
        if ('http://' in href_elem.value) or ('https://' in href_elem.value):
            self.request_uri = href_elem.value
        else:
            self.request_uri = self.base.value + href_elem.value

    async def get_headers(self, forms_elem):
        """
        This method gets the headers for the request from the forms element within the InteractionMetadata element. The
        information is saved in the global variables of the class.

        Args:
            forms_elem (basyx.aas.model.submodelElementCollection): SubmodelElement of forms within InteractionMetadata.
        """
        headers_elem = forms_elem.get_sm_element_by_semantic_id(
            HTTPAssetInterfaceSemantics.SEMANTICID_HTTP_INTERFACE_HEADERS)
        if not headers_elem:
            # This Interaction element does not have headers
            return
        # The old request headers are updated
        self.request_headers = {}
        for header_smc in headers_elem:
            field_name = header_smc.get_sm_element_by_semantic_id(
                HTTPAssetInterfaceSemantics.SEMANTICID_HTTP_INTERFACE_FIELD_NAME).value
            field_value = header_smc.get_sm_element_by_semantic_id(
                HTTPAssetInterfaceSemantics.SEMANTICID_HTTP_INTERFACE_FIELD_VALUE).value
            self.request_headers[field_name] = field_value

    async def get_method_name(self, forms_elem):
        """
        This method gets the method name of the request from the forms element within the InteractionMetadata element.
        The information is saved in the global variables of the class.

        Args:
            forms_elem (basyx.aas.model.submodelElementCollection): SubmodelElement of forms within InteractionMetadata.
        """
        method_name_elem = forms_elem.get_sm_element_by_semantic_id(
            HTTPAssetInterfaceSemantics.SEMANTICID_HTTP_INTERFACE_METHOD_NAME)
        if method_name_elem:
            self.request_method = method_name_elem.value

    async def add_asset_service_data(self, interaction_metadata, service_input_data):
        """
        This method adds the required data of the asset service, using the skill params information (exposure element
        and skill input data). The information is saved in the global variables of the class.

        Args:
            interaction_metadata (basyx.aas.model.SubmodelElementCollection): SubmodelElement of interactionMetadata.
            service_input_data (dict): dictionary containing the input data of the asset service.
        """
        for submodel_element in traversal.walk_submodel(interaction_metadata):
            # TODO AQUI QUEDA COMPROBAR QUE EXISTE EL SEMANTICID PARA DETERMINAR DONDE HAY QUE AÑADIR LOS DATOS (hay que
            #  pensar el nombre, pero algo estilo 'InputDataLocation')
            # TODO DE MOMENTO SE DEJA CON SOLO PARAMS DE HTTP
            if submodel_element.check_semantic_id_exist(HTTPAssetInterfaceSemantics.SEMANTICID_HTTP_INTERFACE_PARAMS):
                request_params = {}
                param_name = submodel_element.get_sm_element_by_semantic_id(
                    HTTPAssetInterfaceSemantics.SEMANTICID_HTTP_INTERFACE_PARAM_NAME).value
                request_params[param_name] = service_input_data[param_name]
                self.request_params = request_params
                break
            # TODO pensar como se haria si no hay que añadirlo en los parametros (p.e. en el body)
            # TODO PENSAR EN MAS OPCIONES DE AÑADIR LOS PARAMETROS (P.E. EN LA URI)
        else:
            AssetConnectionError("The interface need input data but there is no location defined for it.",
                                 'Invalid interface SubmodelElement', 'MissingAttribute')


    async def send_http_request(self):
        """
        This method sends the required HTTP request message to the asset. All the required information is obtained from
         the global variables of the class.

        Returns:
            aiohttp.ClientResponse: response of the asset.
        """
        async with aiohttp.ClientSession(headers=self.request_headers) as session:
            try:
                # async with session.get(url=self.request_uri, params=self.request_params) as resp:
                #     await resp.text()   # The content is saved in ClientResponse object
                #     return resp
                if self.request_method == 'GET':
                    response = await session.get(url=self.request_uri, params=self.request_params)
                elif self.request_method == 'DELETE':
                    # TODO a probar
                    response = await session.delete(url=self.request_uri, params=self.request_params)
                elif self.request_method == 'HEAD':
                    # TODO a probar
                    response = await session.head(url=self.request_uri, params=self.request_params)
                elif self.request_method == 'PATCH':
                    # TODO a probar
                    response = await session.patch(url=self.request_uri, params=self.request_params)
                elif self.request_method == 'POST':
                    # TODO a probar
                    response = await session.post(url=self.request_uri, params=self.request_params)
                elif self.request_method == 'PUT':
                    # TODO a probar
                    response = await session.put(url=self.request_uri, params=self.request_params)
                await response.text()   # The content is saved in ClientResponse object
                return response
            except (ClientConnectorError, ClientConnectionError) as connection_error:
                if isinstance(connection_error, ClientConnectorError):
                    raise AssetConnectionError("The request to asset timed out, so the asset is not available.",
                                               "AssetConnectTimeout", "The asset connection timed out")
                if isinstance(connection_error, ClientConnectionError):
                    raise AssetConnectionError("The connection with the asset has raised an exception.",
                                               connection_error.__class__.__name__, connection_error.args[0].reason)


class HTTPAssetInterfaceSemantics:
    """
    This class contains the specific semanticIDs of HTTP interfaces.
    """

    SEMANTICID_HTTP_INTERFACE_METHOD_NAME = 'https://www.w3.org/2011/http#methodName'
    SEMANTICID_HTTP_INTERFACE_HEADERS = 'https://www.w3.org/2011/http#headers'
    SEMANTICID_HTTP_INTERFACE_FIELD_NAME = 'https://www.w3.org/2011/http#fieldName'
    SEMANTICID_HTTP_INTERFACE_FIELD_VALUE = 'https://www.w3.org/2011/http#fieldValue'

    # TODO nuevo
    SEMANTICID_HTTP_INTERFACE_PARAMS = 'https://www.w3.org/2011/http#params'
    SEMANTICID_HTTP_INTERFACE_PARAM_NAME = 'https://www.w3.org/2011/http#paramName'
    SEMANTICID_HTTP_INTERFACE_PARAM_VALUE = 'https://www.w3.org/2011/http#paramValue'



import abc
import json
import re
from enum import Enum, unique

from jsonpath_ng import parse
from lxml import etree

from smia.logic.exceptions import AssetConnectionError
from smia.utilities.smia_info import AssetInterfacesInfo


class AssetConnection(metaclass=abc.ABCMeta):
    """
    This class is an abstract class for all AssetConnections.
    """

    @unique
    class ArchitectureStyle(Enum):
        PUBSUB = 0
        CLIENTSERVER = 1
        NOTAPPLICABLE = 2

    @abc.abstractmethod
    def __init__(self):
        super().__init__()
        self.architecture_style: AssetConnection.ArchitectureStyle = AssetConnection.ArchitectureStyle.NOTAPPLICABLE

        # General attributes for all Asset Connections
        self.interface_title = None
        self.endpoint_metadata_elem = None

    @abc.abstractmethod
    async def configure_connection_by_aas_model(self, interface_aas_elem):
        """
        This method configures the Asset Connection using the interface element defined in the AAS model.

        Args:
            interface_aas_elem (basyx.aas.model.SubmodelElement): element of the AAS model with the asset interface information.
        """
        pass

    @abc.abstractmethod
    async def check_asset_connection(self):
        """
        This method checks the Asset Connection.
        """
        pass

    @abc.abstractmethod
    async def connect_with_asset(self):
        """
        This method performs the connection process to the Asset Connection.
        """
        pass

    @abc.abstractmethod
    async def execute_asset_service(self, interaction_metadata, service_input_data=None):
        """
        This method sends a message to the asset and returns the response. The connection of the interface of the asset
        is already configured in 'configure_connection_by_aas_model' method, but the interaction metadata is provided
        in form of a Python object of AAS model (SubmodelElement).

        Args:
            interaction_metadata (basyx.aas.model.SubmodelElement): element of the AAS model with all metadata for the interaction with the asset.
            service_input_data: object with the input data of the service

        Returns:
            object: response information defined in the interaction metadata.
        """
        pass

    @abc.abstractmethod
    async def receive_msg_from_asset(self):
        """
        This method receives a message from the asset through the Asset Connection.
        """
        pass

    # ------------------------------------------------------------
    # Useful methods for the configuration of the asset connection
    # ------------------------------------------------------------
    @classmethod
    async def check_interface_element(cls, interface_elem):
        """
        This method checks the given interface AAS element from the 'AssetInterfacesDescription' submodel. In case of
        identifying any error it raises an 'AssetConnectionError' exception.

        Args:
            interface_elem(basyx.aas.model.SubmodelElementCollection): SubmodelElement of AAS interface SubmodelElement.
        """
        # First, it is checked that the Interface element offered is within the Submodel 'AssetInterfacesDescription'.
        parent_submodel = interface_elem.get_parent_submodel()
        if not parent_submodel.check_semantic_id_exist(AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL):
            raise AssetConnectionError("The Interface element object is invalid because it is not within"
                                       " the required submodel", "invalid interface element",
                                       "Interface element is not within the required submodel")
        # The semanticID of the Interface element is also checked
        if not interface_elem.check_semantic_id_exist(AssetInterfacesInfo.SEMANTICID_INTERFACE):
            raise AssetConnectionError("The Interface element object is invalid because it does not have"
                                       " the required semanticID", "invalid interface element",
                                       "Interface element does not have the required semanticID")
        # Then, if required submodel elements are missing is checked
        await cls.check_submodel_element_exist_by_semantic_id(interface_elem, 'title',
                                                              AssetInterfacesInfo.SEMANTICID_INTERFACE_TITLE)
        await cls.check_submodel_element_exist_by_semantic_id(interface_elem, 'EndpointMetadata',
                                                              AssetInterfacesInfo.SEMANTICID_ENDPOINT_METADATA)
        await cls.check_submodel_element_exist_by_semantic_id(interface_elem, 'InteractionMetadata',
                                                              AssetInterfacesInfo.SEMANTICID_INTERACTION_METADATA)
        # TODO comprobar que no haya mas atributos requeridos


    async def check_endpoint_metadata(self):
        """
        This method checks if the given endpointMetadata object is valid (if it is within the correct submodel and if
        it has all required attributes).
        """
        # The endpointMetadata object is initialized in 'configure_connection_by_aas_model' method, but if it is None,
        # then the Interface Element is invalid
        if self.endpoint_metadata_elem is None:
            raise AssetConnectionError("The Interface Element object is invalid because it does not have the"
                                       " EndpointMetadata", "invalid interface",
                                       "EndpointMetadata is not within the Interface Element")

        # It is checked that the metadata SMC offered is within the Submodel 'AssetInterfacesDescription'.
        parent_submodel = self.endpoint_metadata_elem.get_parent_submodel()
        if not parent_submodel.check_semantic_id_exist(AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL):
            raise AssetConnectionError("The EndpointMetadata object is invalid because it is not within"
                                       " the required submodel", "invalid endpoint metadata",
                                       "EndpointMetadata is not within the required submodel")
        # Then, if required submodel elements are missing is checked
        await self.check_submodel_element_exist_by_semantic_id(self.endpoint_metadata_elem, 'base',
                                                              AssetInterfacesInfo.SEMANTICID_INTERFACE_BASE)
        await self.check_submodel_element_exist_by_semantic_id(self.endpoint_metadata_elem, 'contentType',
                                                              AssetInterfacesInfo.SEMANTICID_INTERFACE_CONTENT_TYPE)
        await self.check_submodel_element_exist_by_semantic_id(self.endpoint_metadata_elem, 'securityDefinitions',
                                                              AssetInterfacesInfo.SEMANTICID_INTERFACE_SECURITY_DEFINITIONS)
        # TODO comprobar que no haya mas atributos requeridos

    # -----------------------------------------------------------------
    # Useful methods for the processes prior to connection to the asset
    # -----------------------------------------------------------------
    async def check_interaction_metadata(self, interaction_metadata):
        """
        This method checks if the given interactionMetadata object is valid (if it is within the correct submodel and if
        it has all required attributes).

        Args:
            interaction_metadata (basyx.aas.model.SubmodelElementCollection): SubmodelElement of interactionMetadata.
        """
        # First, it is checked that the metadata SMC offered is within the Submodel 'AssetInterfacesDescription'.
        parent_submodel = interaction_metadata.get_parent_submodel()
        if not parent_submodel.check_semantic_id_exist(AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL):
            raise AssetConnectionError("The InteractionMetadata object is invalid because it is not within"
                                       " the required submodel", "invalid interaction metadata",
                                       "InteractionMetadata is not within the required submodel")
        # Then, if required submodel elements are missing is checked
        await self.check_submodel_element_exist_by_semantic_id(interaction_metadata, 'forms',
                                                              AssetInterfacesInfo.SEMANTICID_INTERFACE_FORMS)
        await self.check_submodel_element_exist_by_semantic_id(
            interaction_metadata.get_sm_element_by_semantic_id(AssetInterfacesInfo.SEMANTICID_INTERFACE_FORMS),
            'href', AssetInterfacesInfo.SEMANTICID_INTERFACE_HREF)
        # TODO comprobar que no haya mas atributos requeridos

        # If dataQuery is defined, the dataQueryType must be appropriate to the content type of the interaction metadata
        data_query_elem = interaction_metadata.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACE_INTERACTION_DATA_QUERY)
        if data_query_elem is not None:
            content_type_elem = await self.get_interaction_metadata_content_type(interaction_metadata)
            await self.check_data_query_type(interaction_metadata.id_short, content_type_elem.value, data_query_elem)

    async def get_interaction_metadata_content_type(self, interaction_metadata):
        """
        This method gets the content type of the interaction metadata. If it is not defined, the type defined in the
        EndpointMetadata will be
        Args:
            interaction_metadata (basyx.aas.model.SubmodelElementCollection): interactionMetadata Python object.

        Returns:
            basyx.aas.model.SubmodelElementElement: Python object of the content type
        """
        forms_elem = interaction_metadata.get_sm_element_by_semantic_id(AssetInterfacesInfo.SEMANTICID_INTERFACE_FORMS)
        content_type_elem = forms_elem.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACE_CONTENT_TYPE)
        if content_type_elem is None:
            # If content type is not defined in the interactionMetada, it should be got from the EndpointMetadata
            content_type_elem = self.endpoint_metadata_elem.get_sm_element_by_semantic_id(
                AssetInterfacesInfo.SEMANTICID_INTERFACE_CONTENT_TYPE)
        return content_type_elem

    @classmethod
    async def check_submodel_element_exist_by_semantic_id(cls, submodel_elem_col, sm_id_short, semantic_id):
        """
        This method checks if a submodelElement with the given semanticID exists within the given
        SubmodelElementCollection.

        Args:
            submodel_elem_col (basyx.aas.model.SubmodelElementCollection): SubmodelElementCollection where the SubmodelElement has to be found.
            sm_id_short (str): idShort of the SubmodelElement to find.
            semantic_id (str): semantic ID to find the required SubmodelElement.
        """
        elem_to_find = submodel_elem_col.get_sm_element_by_semantic_id(semantic_id)
        if not elem_to_find:
            raise AssetConnectionError(f"The {submodel_elem_col.id_short} object is invalid because the given "
                                       f"object does not have required '{sm_id_short}' element",
                                       "missing submodel element",
                                       f"{submodel_elem_col.id_short} does not have required '{sm_id_short}'")

    # --------------------------------------------------------
    # Useful methods to extract information from asset message
    # --------------------------------------------------------
    @classmethod
    async def check_data_query_type(cls, interaction_elem_name, content_type, data_query_elem):
        """
        This method checks if the data query of the interaction metadata element is valid. It is valid when the type
        specified in the data query is the appropriate for the content type of the element.

        Args:
            interaction_elem_name (str): name of the interaction element to show in case of invalid data query.
            content_type (str): type of the content of the interaction metadata element.
            data_query_elem (basyx.aas.model.SubmodelElementCollection): SubmodelElement of dataQuery.

        Returns:

        """
        if data_query_elem is not None:
            data_query_type = data_query_elem.get_qualifier_value_by_type('DataQueryType')
            if (content_type == 'application/json' and data_query_type != 'jsonpath') or \
                (content_type == 'text/plain' and data_query_type != 'regex') or \
                (content_type == 'application/x-www-form-urlencoded' and data_query_type != 'regex') or \
                (content_type == 'application/xml' and data_query_type != 'xpath'):
                        raise AssetConnectionError("The dataQuery type of interaction metadata {} is not valid for "
                                                   "content type {}".format(interaction_elem_name, content_type),
                                                   'invalid data query', "{} does no have a valid data"
                                                                         " query".format(interaction_elem_name))



    @classmethod
    async def get_response_content(cls, interaction_metadata, response_content):
        """
        This method gets the required information from the content of the response message from the asset using the
        interactionMetadata information.

        Args:
            interaction_metadata (basyx.aas.model.SubmodelElementCollection): SubmodelElement of interactionMetadata.
            response_content (str): The content of the response message from the asset in string format.
        """
        # First, if 'dataQuery' attribute is set has to be analyzed
        data_query_elem = interaction_metadata.get_sm_element_by_semantic_id(
            AssetInterfacesInfo.SEMANTICID_INTERFACE_INTERACTION_DATA_QUERY)
        if data_query_elem is not None:
            # The type of the content must be obtained
            forms_elem = interaction_metadata.get_sm_element_by_semantic_id(
                AssetInterfacesInfo.SEMANTICID_INTERFACE_FORMS)
            content_type = forms_elem.get_sm_element_by_semantic_id(
                AssetInterfacesInfo.SEMANTICID_INTERFACE_CONTENT_TYPE).value
            # The general method for all Asset Connections will be used
            response_content = await cls.extract_information_with_data_query(content_type, response_content,
                                                                             data_query_elem.value)
            return response_content
        # If the type of the interaction element of the interface is added, it needs a transformation of data type
        data_type = interaction_metadata.get_sm_element_by_semantic_id(
                AssetInterfacesInfo.SEMANTICID_INTERFACE_INTERACTION_TYPE)
        if data_type is not None:
            return await cls.transform_data_by_type(response_content, data_type.value)
        else:
            return response_content

    @classmethod
    async def extract_information_with_data_query(cls, content_type, content_data, query):
        """
        This method extracts the information of the content using the query defined in the interface element.

        Args:
            content_type (str): format of the content.
            content_data: data of the content.
            query (str): query to extract information from the content.

        Returns:
             object: extracted information (result of the query in the given data).
        """
        if 'application/json' in content_type:
            return cls.extract_from_json(content_data, query)
        elif 'application/xml' in content_type:
            return cls.extract_from_xml(content_data, query)
        elif ('text/plain' in content_type) or ('application/x-www-form-urlencoded' in content_type):
            return cls.extract_from_string(content_data, query)
        else:
            raise ValueError(f'Unsupported Content-Type: {content_type}')

    @classmethod
    async def transform_data_by_type(cls, content_data, data_type):
        """
        This method transforms the data by its type.

        Args:
            content_data: data of the content to be transformed.
            data_type: type of the data.

        Returns:
            object: transformed data.
        """
        match data_type:
            case 'string':
                return str(content_data)
            case 'number':
                return float(content_data)
            case 'integer':
                return int(content_data)
            case 'boolean':
                return bool(content_data)
            case 'object':
                if isinstance(content_data, str):
                    return json.loads(content_data)
                return content_data
            case _:
                return content_data

    @classmethod
    def extract_from_json(cls, data, query):
        """
        This method extracts the required information from a JSON object. It uses the 'jsonpath-ng' Python package to
        perform the extraction.

        Args:
            data (dict): JSON object with the given data.
            query (str): query to extract information from the content. In this case, it has to be a JSONPath expression.

        Returns:
            object: extracted information (result of the query in the given data).
        """
        jsonpath_expr = parse(query)
        extracted_data = [match.value for match in jsonpath_expr.find(json.loads(data))]
        return extracted_data[0] if len(extracted_data) == 1 else extracted_data

    @classmethod
    def extract_from_xml(cls, data, query):
        """
        This method extracts the required information from an XML object. It uses the 'lxml' Python package to
        perform the extraction.

        Args:
            data (str): XML object with the given data in string format.
            query (str): query to extract information from the content. In this case, it has to be a XPath expression.

        Returns:
            object: extracted information (result of the query in the given data).
        """
        root = etree.fromstring(data)
        extracted_data = root.xpath(query)
        return extracted_data[0] if len(extracted_data) == 1 else extracted_data

    @classmethod
    def extract_from_string(cls, data, query):
        """
        This method extracts the required information from a string object. It uses the 're' Python package to
        perform the extraction.

        Args:
            data (str): string object with the given data.
            query (str): query to extract information from the content. In this case, it has to be a Regex expression.

        Returns:
            object: extracted information (result of the query in the given data).
        """
        regex_pattern = re.compile(query)
        extracted_data = regex_pattern.findall(data)
        return extracted_data[0] if len(extracted_data)==1 else extracted_data



import logging

from spade.behaviour import State

from smia.utilities import smia_archive_utils

_logger = logging.getLogger(__name__)


class StateStopping(State):
    """
    This class contains the Stop state of the common SMIA.
    """

    # TODO sin acabar

    async def run(self):
        """
        This method implements the stop state of the common SMIA.
        """
        _logger.info("## STATE 3: STOPPING ##")
        # sb = StoppingBehaviour(self.agent)
        # self.agent.add_behaviour(sb)
        _logger.info("     [Stopping Behaviour]")
        _logger.info("         |___ Stopping MachineAgent...")
        smia_archive_utils.update_status('Stopping')
        # No se ha indicado un estado final, por lo que este se considera el último



import logging

from basyx.aas.adapter import aasx
from owlready2 import OneOf

from smia import SMIAGeneralInfo
from smia.logic.exceptions import OntologyReadingError, CriticalError
from smia.utilities import properties_file_utils, smia_archive_utils

_logger = logging.getLogger(__name__)


class CapabilitySkillOntologyUtils:
    """This class contains all information about the proposal of the ontology based on Capability-Skill model. This
    information groups the required semanticIDs or the qualifiers to analyze AAS models."""

    @staticmethod
    def get_ontology_file_path():
        """
        This method gets the valid file path of the ontology file. The file is obtained from the AASX package or from
        the configuration folder inside the SMIA Archive, depending on the definition in the properties file.

        Returns:
            str: file path to the ontology file.
        """
        ontology_property = properties_file_utils.get_ontology_general_property("inside-aasx")
        if (ontology_property == '#') or (ontology_property.lower() in ('yes', 'true', 't', '1')):
            # The ontology file will be checked if the ontology file is inside the AASX package if the initialization
            # properties file is not added ('inside-aasx' property is '#') or is added and set as such.
            aas_model_serialization_format = properties_file_utils.get_aas_general_property('model.serialization')
            if aas_model_serialization_format == '.aasx':
                try:
                    with aasx.AASXReader(properties_file_utils.get_aas_model_filepath()) as aasx_reader:
                        for part_name, content_type in aasx_reader.reader.list_parts():
                            # All parts of the AASX package are analyzed
                            # if '.owl' in part_name:
                            if any(ext in part_name for ext in ['.owl', '.rdf', '.owx']):
                                ontology_zip_file = aasx_reader.reader.open_part(part_name)
                                ontology_file_bytes = ontology_zip_file.read()
                                return properties_file_utils.create_ontology_file(ontology_file_bytes)
                        else:
                            # If the ontology file is not inside the AASX and is not defined in the initialization
                            # properties file, an OWL file will be looked for inside the SMIA Archive.
                            _logger.info("The ontology OWL file is not found within the AASX package, so it will be "
                                         "searched for within the SMIA Archive.")
                            ontology_file_path = smia_archive_utils.get_file_by_extension(
                                SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH, '.owl')
                            if ontology_file_path is None:
                                raise CriticalError("Failed to read ontology file inside the AASX package (FileNotFound)"
                                                    " and inside the SMIA Archive.")
                            else:
                                return ontology_file_path

                except ValueError as e:
                    raise CriticalError("Failed to read AAS model: invalid file.")
            else:
                # The serialization format must be AASX. However, it will be checked if the ontology file defined in
                # the properties file is valid.
                _logger.warning("The properties file may not be well defined. The ontology file is set as is within "
                                "AASX, but the serialization format of the AAS model is not AASX.")
                # It will try with the file path defined in the properties file, if it does not exist, it will crash
                # during the loading of the ontology
                return properties_file_utils.get_defined_ontology_filepath()
        else:
            return properties_file_utils.get_defined_ontology_filepath()

    @staticmethod
    def get_possible_values_of_datatype(datatype):
        """
        This method returns all possible values of an OWL data type. If the data type does not have the equivalent
        'OneOf', so the values do not need to be constrained and validated, it returns None.

        Args:
            datatype (owlready2.Oneof): OWL datatype object.

        Returns:
            set: possible values of datatype in form of a list of strings.
        """
        possible_values = set()
        if datatype.equivalent_to:  # Check for equivalent classes
            for equivalent in datatype.equivalent_to:
                if isinstance(equivalent, OneOf):
                    for value in equivalent.instances:
                        possible_values.add(str(value))
        if len(possible_values) == 0:
            return None
        return possible_values

    @staticmethod
    def check_and_get_xsd_datatypes(datatype):
        """
        This method checks whether the given OWL data type is one of those defined in XSD and, if true, returns the
        associated data type. If false, it returns None.

        Args:
            datatype (owlready2.Oneof): OWL datatype object.

        Returns:
            object: value of datatype defined in XSD (None if it is not found).
        """
        if datatype.equivalent_to:  # Check for equivalent classes
            for equivalent in datatype.equivalent_to:
                if equivalent == str:
                    return str
            return None

    @staticmethod
    def check_whether_part_of_domain(owl_instance, domain):
        """
        This method checks whether a given instance class is part of a given domain.

        Args:
            owl_instance (ThingClass): instance of the OWL class to be checked.
            domain (CallbackList): list of all classes within the given domain.

        Returns:
            bool: result of the check.
        """
        for domain_class in domain:
            for owl_class in owl_instance.is_a:
                if owl_class == domain_class or domain_class in owl_class.ancestors():
                    return True
        return False

    @staticmethod
    def get_aas_classes_from_object_property(object_property_class):
        """
        This method gets the AAS class related to the of the ontology classes of a given Object Property. If the
        attribute of the AAS class does not exist, it raises an exception.

        Args:
            object_property_class (ObjectPropertyClass): class object of the ObjectProperty.

        Returns:
            object, object: AAS class value of the domain and range ontology classes.
        """
        if object_property_class is None:
            raise OntologyReadingError("The object property object {} is None".format(object_property_class))
        if len(object_property_class.domain) == 0 or len(object_property_class.range) == 0:
            raise OntologyReadingError("The domain or range of object property {} are "
                                       "empty".format(object_property_class))
        try:
            for domain_class in object_property_class.domain:
                if hasattr(domain_class, 'get_associated_aas_class'):
                    domain_aas_class = domain_class.get_associated_aas_class()
                    break
            else:
                # In this case no domain class is of the defined CSS model of the SMIA approach.
                raise OntologyReadingError("The domain of object property object {} does not have any associated "
                                           "AAS model classes defined".format(object_property_class))
            for range_class in object_property_class.range:
                if hasattr(range_class, 'get_associated_aas_class'):
                    range_aas_class = range_class.get_associated_aas_class()
                    break
            else:
                # In this case no range class is of the defined CSS model of the SMIA approach.
                raise OntologyReadingError("The range of object property object {} does not have any associated "
                                           "AAS model classes defined".format(object_property_class))
            return domain_aas_class, range_aas_class
        except KeyError as e:
            raise OntologyReadingError("The domain or range of object property object {} does not have associated "
                                           "AAS model classes defined".format(object_property_class))

    # Types of Capabilities
    MANUFACTURING_CAPABILITY_TYPE = 'ManufacturingCapability'
    ASSET_CAPABILITY_TYPE = 'AssetCapability'
    AGENT_CAPABILITY_TYPE = 'AgentCapability'
    CAPABILITY_TYPE_POSSIBLE_VALUES = [MANUFACTURING_CAPABILITY_TYPE, ASSET_CAPABILITY_TYPE, AGENT_CAPABILITY_TYPE]

    # SemanticIDs of Capabilities
    # TODO ESTA LA DEJO PORQUE NO SE HA UTILIZADO TODAVIA EN LA ONTOLOGIA, PERO TAMBIEN HAY QUE ELIMINARLA
    SEMANTICID_REL_SKILL_PARAMETER_SKILL_INTERFACE = 'urn:ehu:gcis:capabilityskillontology:1:1:skillparameterexposedthrough'


    # Qualifiers for Capabilities
    QUALIFIER_CAPABILITY_TYPE = 'ExpressionSemantic'
    QUALIFIER_CAPABILITY_POSSIBLE_VALUES = ['REQUIREMENT', 'OFFER', 'ASSURANCE']

    # Qualifiers for Skills
    QUALIFIER_SKILL_TYPE = 'SkillImplementationType'
    QUALIFIER_SKILL_POSSIBLE_VALUES = ['STATE', 'TRIGGER', 'OPERATION', 'FUNCTIONBLOCK']

    # Qualifiers for Feasibility Checking
    QUALIFIER_FEASIBILITY_CHECKING_TYPE = 'FeasibilityCheckingCondition'
    QUALIFIER_FEASIBILITY_CHECKING_POSSIBLE_VALUES = ['PRECONDITION', 'INVARIANT', 'POSTCONDITION']

    # IDs for Negotiation AgentCapability
    CONCEPT_DESCRIPTION_ID_NEGOTIATION_CRITERIA = 'urn:ehu:gcis:conceptdescriptions:1:1:negotiationcriteria'


class CapabilitySkillOntologyInfo:
    """
    This class contains information related to the ontology of Capability-Skill: namespaces, OWL file...
    """
    CSS_ONTOLOGY_BASE_NAMESPACE = 'http://www.w3id.org/hsu-aut/css#'
    CSS_ONTOLOGY_SMIA_NAMESPACE = 'http://www.w3id.org/upv-ehu/gcis/css-smia#'

    # SemanticIDs (IRIs) of Capabilities
    CSS_ONTOLOGY_CAPABILITY_IRI = 'http://www.w3id.org/hsu-aut/css#Capability'
    CSS_ONTOLOGY_AGENT_CAPABILITY_IRI = 'http://www.w3id.org/upv-ehu/gcis/css-smia#AgentCapability'
    CSS_ONTOLOGY_ASSET_CAPABILITY_IRI = 'http://www.w3id.org/upv-ehu/gcis/css-smia#AssetCapability'
    CSS_ONTOLOGY_CAPABILITY_CONSTRAINT_IRI = 'http://www.w3id.org/hsu-aut/css#CapabilityConstraint'

    # SemanticIDs (IRIs) of Skills
    CSS_ONTOLOGY_SKILL_IRI = 'http://www.w3id.org/hsu-aut/css#Skill'
    CSS_ONTOLOGY_SKILL_INTERFACE_IRI = 'http://www.w3id.org/hsu-aut/css#SkillInterface'
    CSS_ONTOLOGY_SKILL_PARAMETER_IRI = 'http://www.w3id.org/hsu-aut/css#SkillParameter'
    CSS_ONTOLOGY_SKILL_STATE_MACHINE_IRI = 'http://www.w3id.org/hsu-aut/css#StateMachine'

    # SemanticIDs (IRIs) of Object Properties (relationship between classes)
    CSS_ONTOLOGY_PROP_ISREALIZEDBY_IRI = 'http://www.w3id.org/hsu-aut/css#isRealizedBy'
    CSS_ONTOLOGY_PROP_ISRESTRICTEDBY_IRI = 'http://www.w3id.org/hsu-aut/css#isRestrictedBy'
    CSS_ONTOLOGY_PROP_ACCESSIBLETHROUGH_IRI = 'http://www.w3id.org/hsu-aut/css#accessibleThrough'
    CSS_ONTOLOGY_PROP_ACCESSIBLETHROUGH_ASSET_IRI = 'http://www.w3id.org/upv-ehu/gcis/css-smia#accessibleThroughAssetService'
    CSS_ONTOLOGY_PROP_ACCESSIBLETHROUGH_AGENT_IRI = 'http://www.w3id.org/upv-ehu/gcis/css-smia#accessibleThroughAgentService'
    CSS_ONTOLOGY_PROP_HASPARAMETER_IRI = 'http://www.w3id.org/hsu-aut/css#hasParameter'
    CSS_ONTOLOGY_PROP_BEHAVIOURSCONFORMSTO_IRI = 'http://www.w3id.org/hsu-aut/css#behaviorConformsTo'

    CSS_ONTOLOGY_THING_CLASSES_IRIS = [CSS_ONTOLOGY_CAPABILITY_IRI,
                                       CSS_ONTOLOGY_AGENT_CAPABILITY_IRI,
                                       CSS_ONTOLOGY_ASSET_CAPABILITY_IRI,
                                       CSS_ONTOLOGY_CAPABILITY_CONSTRAINT_IRI,
                                       CSS_ONTOLOGY_SKILL_IRI,
                                       CSS_ONTOLOGY_SKILL_INTERFACE_IRI,
                                       CSS_ONTOLOGY_SKILL_PARAMETER_IRI,
                                       CSS_ONTOLOGY_SKILL_STATE_MACHINE_IRI,
                                       ]

    CSS_ONTOLOGY_OBJECT_PROPERTIES_IRIS = [CSS_ONTOLOGY_PROP_ISREALIZEDBY_IRI,
                                           CSS_ONTOLOGY_PROP_ISRESTRICTEDBY_IRI,
                                           CSS_ONTOLOGY_PROP_ACCESSIBLETHROUGH_IRI,
                                           CSS_ONTOLOGY_PROP_ACCESSIBLETHROUGH_ASSET_IRI,
                                           CSS_ONTOLOGY_PROP_ACCESSIBLETHROUGH_AGENT_IRI,
                                           CSS_ONTOLOGY_PROP_HASPARAMETER_IRI,
                                           CSS_ONTOLOGY_PROP_BEHAVIOURSCONFORMSTO_IRI
                                           ]


class CSSModelAASModelInfo:
    """
    This class contains information about the relation between the Capability-Skill-Service model and the AAS model.
    """
    from smia.aas_model import extended_submodel

    CSS_ONTOLOGY_AAS_MODEL_LINK = {
        CapabilitySkillOntologyInfo.CSS_ONTOLOGY_CAPABILITY_IRI: extended_submodel.ExtendedCapability,
        CapabilitySkillOntologyInfo.CSS_ONTOLOGY_AGENT_CAPABILITY_IRI: extended_submodel.ExtendedCapability,
        CapabilitySkillOntologyInfo.CSS_ONTOLOGY_ASSET_CAPABILITY_IRI: extended_submodel.ExtendedCapability,
        CapabilitySkillOntologyInfo.CSS_ONTOLOGY_CAPABILITY_CONSTRAINT_IRI: extended_submodel.ExtendedCapabilityConstraint,
        CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_IRI: extended_submodel.ExtendedSkill,
        CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_INTERFACE_IRI: extended_submodel.ExtendedSkillInterface,
        CapabilitySkillOntologyInfo.CSS_ONTOLOGY_SKILL_PARAMETER_IRI: extended_submodel.ExtendedSkillParameter,
        # CSS_ONTOLOGY_SKILL_STATE_MACHINE_IRI: '',     # TODO todavia no se ha usado
    }


class CapabilitySkillACLInfo:
    """
    This subclass of CapabilitySkillOntology contains information about the structure of the message that are sent
    between DTs in relation to Capability-Skill model.
    """

    # Required Capability information
    REQUIRED_CAPABILITY_NAME = 'capabilityName'
    REQUIRED_CAPABILITY_TYPE = 'capabilityType'
    REQUIRED_CAPABILITY_CONSTRAINTS = 'capabilityConstraints'
    REQUIRED_SKILL_INFO = 'skillInfo'
    REQUIRED_SKILL_NAME = 'skillName'
    REQUIRED_SKILL_ELEMENT_TYPE = 'smeType'
    REQUIRED_SKILL_PARAMETERS = 'skillParameters'
    REQUIRED_SKILL_PARAMETERS_VALUES = 'skillParameterValues'
    REQUIRED_SKILL_INPUT_PARAMETERS = 'inputs'
    REQUIRED_SKILL_OUTPUT_PARAMETERS = 'outputs'
    REQUIRED_SKILL_INTERFACE_NAME = 'skillInterfaceName'
    # TODO pensar si harian falta mas





"""
This file contains all the classes for handling errors in exceptions that may occur during code execution.
"""
import logging

from smia.utilities.fipa_acl_info import FIPAACLInfo, ServiceTypes

_logger = logging.getLogger(__name__)


class CriticalError(Exception):
    """
    This exception class is defined for errors that are critical to the program, so that execution must be terminated.
    """

    def __init__(self, message):
        _logger.error(f"{message}")
        _logger.error("The program must end.")
        exit(-1)

class RequestDataError(Exception):
    """
    This exception class is defined for errors that are related to a request received by the DT (request of a service
    or a capability).
    """

    def __init__(self, message):
        self.message = message
        _logger.error(f"{self.message}")

# AAS model-related exceptions
# ----------------------------------------
class AASModelReadingError(Exception):
    """
    This exception class is defined for errors that occur during AAS model management (reading, updating...).
    """
    def __init__(self, message, sme_class, reason):
        self.message = message
        self.sme_class = sme_class
        self.reason = reason
        _logger.error(f"{self.message}")


class AASModelOntologyError(Exception):
    """
    This exception class is defined for errors that occur during AAS model management (reading, updating...).
    """
    def __init__(self, message, sme_class, reason):
        self.message = message
        self.sme_class = sme_class
        self.reason = reason
        _logger.error(f"{self.message}")


# Service management-related exceptions
# -------------------------------------
class ServiceRequestExecutionError(Exception):
    """
    This exception class is defined for errors that are related to the execution of a requested service. Since it has
    been requested, this class also must response to the requester with a Failure of the execution.
    """

    def __init__(self, thread, message, svc_type, behav_class):
        self.thread = thread
        self.message = message
        self.svc_type = svc_type
        self.behav_class = behav_class

    async def handle_service_execution_error(self):
        """
        This method handles the error during an execution of a service, sending the Failure message to the requester.
        """
        _logger.error(f"{self.message}")

        _logger.info("Due to an incorrect execution of the service related to the thread [{}], the requester shall be "
                     "informed with a Failure message.".format(self.thread))

        await self.behav_class.send_response_msg_to_sender(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE,
                                                           {'reason': self.message})
        _logger.info("Failure message sent to the requester related to the thread [{}].".format(self.thread))

        # The information about the error is also saved in the log
        from smia.behaviours.specific_handle_behaviours.handle_svc_request_behaviour import HandleSvcRequestBehaviour
        from smia.behaviours.specific_handle_behaviours.handle_capability_behaviour import HandleCapabilityBehaviour
        from smia import GeneralUtils  # Local imports to avoid circular import error
        from smia.utilities import smia_archive_utils

        if isinstance(self.behav_class, (HandleSvcRequestBehaviour, HandleCapabilityBehaviour)):
            acl_info = self.behav_class.svc_req_data
        else:
            acl_info = {'thread': self.thread}
        smia_archive_utils.save_svc_error_log_info(GeneralUtils.get_current_timestamp(), acl_info, self.message,
                                                   self.svc_type)

        # The behaviour for the execution of the service must be killed
        self.behav_class.kill(exit_code=10)

# Capability management-related exceptions
# ----------------------------------------
class CapabilityRequestExecutionError(Exception):
    """
    This exception class is defined for errors that are related to the execution of a requested Capability. Since it has
    been requested, this class also must response to the requester with a Failure of the capability execution.
    """

    def __init__(self, cap_name, message, behav_class):
        self.cap_name = cap_name  # TODO pensar si en lugar de nombre de capacidad añadir el thread de la conversacion (quizas es mas identificativo)
        self.message = message
        self.behav_class = behav_class

    async def handle_capability_execution_error(self):
        """
        This method handles the error during an execution of a capability, sending the Failure message to the requester.
        """
        _logger.error(f"{self.message}")

        _logger.info("Due to an incorrect execution of the capability [{}], the requester shall be informed with a "
                     "Failure message.".format(self.cap_name))

        await self.behav_class.send_response_msg_to_sender(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE,
                                                           {'reason': self.message})
        _logger.info("Failure message sent to the requester of the capability [{}].".format(self.cap_name))

        # The information about the error is also saved in the log
        from smia.behaviours.specific_handle_behaviours.handle_svc_request_behaviour import HandleSvcRequestBehaviour
        from smia.behaviours.specific_handle_behaviours.handle_capability_behaviour import HandleCapabilityBehaviour
        from smia.behaviours.specific_handle_behaviours.handle_negotiation_behaviour import HandleNegotiationBehaviour
        from smia.utilities import smia_archive_utils
        from smia import GeneralUtils  # Local imports to avoid circular import error

        if isinstance(self.behav_class, (HandleSvcRequestBehaviour, HandleCapabilityBehaviour,
                                         HandleNegotiationBehaviour)):
            acl_info = self.behav_class.svc_req_data
        else:
            acl_info = {}
        smia_archive_utils.save_svc_error_log_info(GeneralUtils.get_current_timestamp(), acl_info, self.message,
                                                   ServiceTypes.CSS_RELATED_SERVICE)

        # The behaviour for the execution of the capability must be killed
        self.behav_class.kill(exit_code=10)


class CapabilityCheckingError(Exception):
    """
    This exception class is defined for errors that are related to the execution of the Capability checking process.
    Since this process is requested by another DT, this class also must response to the requester with a Failure of the
    capability checking as well as the reason of it.
    """

    def __init__(self, cap_name, reason, behav_class):
        self.cap_name = cap_name
        self.reason = reason
        self.behav_class = behav_class

    async def handle_capability_checking_error(self):
        """
        This method handles the error during an execution of a capability, sending the Failure message to the requester.
        """
        _logger.error(f"Capability checking of capability [{self.cap_name}] failed. Reason: {self.reason}")

        _logger.info("Due to an incorrect checking of a capability, the requester shall be informed with a Failure "
                     "message.")

        await self.behav_class.send_response_msg_to_sender(
            FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE,
            {'message': "Capability checking of capability [{}] failed".format(self.cap_name), 'reason': self.reason})
        _logger.info("Failure message sent to the requester of the capability.")

        # TODO Pensar si añadir un objeto global en el agente para almacenar informacion sobre errores
        # The behaviour for the execution of the capability must be killed
        self.behav_class.kill(exit_code=10)


class AssetConnectionError(Exception):
    """
    This exception class is defined for errors that are related to the asset connection processes.
    """

    def __init__(self, message, error_type, reason):
        self.message = message
        self.error_type = error_type
        self.reason = reason
        _logger.error(f"{self.message}")

# Capability-Skill ontology exceptions
# ------------------------------------
class OntologyInstanceCreationError(Exception):
    def __init__(self, message):
        self.message = message
        _logger.error(f"{self.message}")


class OntologyReadingError(Exception):
    def __init__(self, message):
        self.message = message
        _logger.error(f"{self.message}")


class OntologyCheckingAttributeError(Exception):

    def __init__(self, message, invalid_instance):
        self.message = message
        self.invalid_instance = invalid_instance
        _logger.error(f"{self.message}")


class OntologyCheckingPropertyError(Exception):

    def __init__(self, message, concerned_property_name, invalid_instance):
        self.message = message
        self.concerned_property_name = concerned_property_name
        self.invalid_instance = invalid_instance
        _logger.error(f"{self.message}")



import asyncio
import logging

import basyx.aas.model
from spade.behaviour import CyclicBehaviour, OneShotBehaviour, TimeoutBehaviour, PeriodicBehaviour, FSMBehaviour

from smia import AASModelUtils
from smia.agents.smia_agent import SMIAAgent
from smia.assetconnection.asset_connection import AssetConnection
from smia.utilities.smia_info import AssetInterfacesInfo

_logger = logging.getLogger(__name__)


class ExtensibleSMIAAgent(SMIAAgent):
    """
    This agent offers some extension mechanisms to add own code to the base :term:`SMIA` agent.
    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.extended_agent_capabilities = []

    # TODO The features that shall be offered are:
    #  1. define and add new connections to the asset (i.e., new communication protocols)
    #  2. define and add new agent capabilities to increase its intelligence and autonomy.
    #  3. define and add new agent services

    def add_new_agent_capability(self, behaviour_class):
        """
        This method adds a new :term:`agent capability <Capability>` to SMIA to increase its intelligence and autonomy.
        The new capability is added as a SPADE behavior instance. If it is not a valid behavior (within the types
        offered by :term`SPADE`), it cannot be added.

        Args:
            behaviour_class: SPADE behaviour class instance.
        """
        if isinstance(behaviour_class, (CyclicBehaviour, PeriodicBehaviour, OneShotBehaviour, TimeoutBehaviour,
                                        FSMBehaviour)):
            self.extended_agent_capabilities.append(behaviour_class)
            _logger.info("Added a new agent capability to SMIA: {}".format(behaviour_class))
        else:
            _logger.warning("The new agent capability [{}] cannot be added because it is not a SPADE behavior "
                            "class.".format(behaviour_class))

    def add_new_agent_service(self, service_id, service_method):
        """
        This method adds a new agent service to SMIA to increase its intelligence and autonomy. The new service is added
         as a Python method that will be called when the service is requested.

        Args:
            service_id (str): identifier of the new service (id or idShort of the related AAS SubmodelElement).
            service_method: Python method that will be called when the service is requested.
        """
        if (service_id is None) or (service_method is None):
            _logger.warning("The new agent service cannot be added because the service identifier or method were not "
                            "provided.")
        else:
            asyncio.run(self.agent_services.save_agent_service(service_id, service_method))
            _logger.info("Added a new agent service to SMIA: {}".format(service_id))

    def add_new_asset_connection(self, aas_interface_id_short, asset_connection_class):
        """
        This method adds a new asset connection to SMIA. The new connection is added by the instance class inherited
        from the official SMIA generic class named 'AssetConnection' and the associated AAS interface element. To
        correctly perform the addition, make sure that the given instance is inherited from this class and that the
        idShort represents the valid AAS SubmodelElement of the related interface within the
        'AssetInterfacesDescription' submodel .

        Args:
            aas_interface_id_short (str): identifier of the related AAS interface element in the form of idshort of the SubmodelElement.
            asset_connection_class: instance class to be added as a new asset connection (inheriting from 'AssetConnection').
        """
        if not isinstance(asset_connection_class, AssetConnection):
            _logger.warning("The new asset connection [{}] cannot be added because it does not inherit from the "
                            "official SMIA class 'AssetConnection'.".format(asset_connection_class))
            return
        if aas_interface_id_short is None:
            _logger.warning("The new asset connection [{}] cannot be added because the idShort of the related AAS "
                            "interface element was not provided.".format(asset_connection_class))
            return

        async def check_and_save_asset_connection():
            object_store = AASModelUtils.read_aas_model_object_store()
            await self.aas_model.set_aas_model_object_store(object_store)
            aid_submodel = await self.aas_model.get_submodel_by_semantic_id(
                AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL)
            if (aid_submodel is None) or (not isinstance(aid_submodel, basyx.aas.model.Submodel)):
                _logger.warning("The standardized submodel 'AssetInterfacesDescription' does not exist in the AAS "
                                "model.".format(asset_connection_class))
                return
            try:
                aas_interface_elem = aid_submodel.get_referable(aas_interface_id_short)
                aas_interface_elem_ref = basyx.aas.model.ModelReference.from_referable(aas_interface_elem)

                # At this point, all checks have been passed, so the new asset connection can be added.
                await self.save_asset_connection_class(aas_interface_elem_ref, asset_connection_class)
                self.conns = {aas_interface_elem_ref: asset_connection_class}
            except KeyError as e:
                _logger.warning(e)
                _logger.warning("The new asset connection [{}] cannot be added because the idShort of the related AAS "
                                "interface element is not valid.".format(asset_connection_class))
                return

        asyncio.run(check_and_save_asset_connection())
        _logger.info("Added a new asset connection to SMIA: {}".format(aas_interface_id_short))




import logging
from os import path

import basyx
from basyx.aas import model
from basyx.aas.adapter import aasx
from basyx.aas.util import traversal

from smia.logic.exceptions import CriticalError, AASModelReadingError
from smia.utilities import properties_file_utils
from smia.utilities.smia_general_info import SMIAGeneralInfo

_logger = logging.getLogger(__name__)


class AASModelUtils:
    """This class contains utility methods related to the AAS model."""

    @staticmethod
    def read_aas_model_object_store():
        """
        This method reads the AAS model according to the selected serialization format.

        Returns:
            basyx.aas.model.DictObjectStore:  object with all Python elements of the AAS model.
        """
        object_store = None
        aas_model_file_path = properties_file_utils.get_aas_model_filepath()
        aas_model_file_name, aas_model_file_extension = path.splitext(SMIAGeneralInfo.CM_AAS_MODEL_FILENAME)
        try:
            # The AAS model is read depending on the serialization format (extension of the AAS model file)
            if aas_model_file_extension == '.json':
                object_store = basyx.aas.adapter.json.read_aas_json_file(aas_model_file_path)
            elif aas_model_file_extension == '.xml':
                object_store = basyx.aas.adapter.xml.read_aas_xml_file(aas_model_file_path)
            elif aas_model_file_extension == '.aasx':
                with aasx.AASXReader(aas_model_file_path) as reader:
                    # Read all contained AAS objects and all referenced auxiliary files
                    object_store = model.DictObjectStore()
                    suppl_file_store = aasx.DictSupplementaryFileContainer()
                    reader.read_into(object_store=object_store,
                                     file_store=suppl_file_store)
        except ValueError as e:
            _logger.error("Failed to read AAS model: invalid file.")
            _logger.error(e)
            raise CriticalError("Failed to read AAS model: invalid file.")
        if object_store is None or len(object_store) == 0:
            raise CriticalError("The AAS model is not valid. It is not possible to read and obtain elements of the AAS "
                                "metamodel.")
        else:
            return object_store

    # Methods related to AASX Package
    # -------------------------------
    @staticmethod
    def get_file_bytes_from_aasx_by_path(file_path):
        """
        This method gets the bytes of a file appended inside an AASX package by a given internal path.

        Args:
            file_path (str): the internal path of the file within the AASX package.

        Returns:
            obj: the content of the file in the form of bytes.
        """
        with aasx.AASXReader(properties_file_utils.get_aas_model_filepath()) as aasx_reader:
            for part_name, content_type in aasx_reader.reader.list_parts():
                if part_name == file_path:
                    return aasx_reader.reader.open_part(part_name).read()
            else:
                _logger.warning("The file with path {} does not find within the AASX Package.".format(file_path))
            return None

    # Methods related to the AAS model
    # --------------------------------
    @staticmethod
    def get_configuration_file_path_from_standard_submodel():
        """
        This method gets the configuration file defined in the 'Software Nameplate' submodel, used as standard submodel
        for the SMIA software definition.

        Returns:
            str: path inside the AASX package of the configuration file.
        """
        # First, the AAS model need to be read
        object_store = AASModelUtils.read_aas_model_object_store()
        soft_nameplate_config_paths = AASModelUtils.get_elem_of_software_nameplate_by_semantic_id(object_store)
        if soft_nameplate_config_paths is None:
            raise CriticalError(
                "Configuration of SMIA is required and it is not defined within the Software Nameplate submodel.")
        for config_path_elem in soft_nameplate_config_paths:
            sm_elem = config_path_elem.get_sm_element_by_semantic_id(
                AASModelInfo.SEMANTIC_ID_SOFTWARE_NAMEPLATE_CONFIG_TYPE)
            if (sm_elem is not None) and (sm_elem.value == 'initial configuration'):
                return config_path_elem.get_sm_element_by_semantic_id(
                    AASModelInfo.SEMANTIC_ID_SOFTWARE_NAMEPLATE_CONFIG_URI).value
        return None

    @staticmethod
    def get_elem_of_software_nameplate_by_semantic_id(object_store):
        """
        This method obtains a SubmodelElement of the Software Nameplate submodel.

        Args:
            object_store (basyx.aas.model.DictObjectStore): storage with all AAS information,

        Returns:
            basyx.aas.model.SubmodelElement: submodelElement with the given semanticID
        """
        for object in object_store:
            if isinstance(object, basyx.aas.model.Submodel):
                if object.check_semantic_id_exist(AASModelInfo.SEMANTICID_SOFTWARE_NAMEPLATE_SUBMODEL):
                    for elem in traversal.walk_submodel(object):
                        if isinstance(elem, basyx.aas.model.SubmodelElement):
                            if elem.check_semantic_id_exist(AASModelInfo.SEMANTIC_ID_SOFTWARE_NAMEPLATE_CONFIG_PATHS):
                                return elem
        return None

    @staticmethod
    async def get_key_type_by_string(key_type_string):
        """
        This method gets the KeyType defined in BaSyx SDK related to the given string.

        Args:
            key_type_string (str): string of the desired KeyType.

        Returns:
            basyx.aas.model.KeyTypes: object of the KeyType defined in BaSyx.
        """
        try:
            return getattr(basyx.aas.model.KeyTypes, key_type_string)
        except AttributeError as e:
            _logger.error(e)
            raise AASModelReadingError("The KeyType with string {} does not exist in the AAS model defined"
                                       " types".format(key_type_string), sme_class=None, reason='KeyTypeAttributeError')

    @staticmethod
    async def get_model_type_by_key_type(key_type):
        """
        This method gets the AAS model class by a given KeyType defined in BaSyx SDK.

        Args:
            key_type (basyx.aas.model.KeyTypes): desired KeyType.

        Returns:
            object of the AAS model class.
        """
        for model_class, key_type_class in basyx.aas.model.KEY_TYPES_CLASSES.items():
            if key_type_class == key_type:
                return model_class
        return None

    @staticmethod
    async def create_aas_reference_object(reference_type, keys_dict=None, external_ref=None):
        """
        This method creates the AAS BaSyx Reference Python object. Depending on the reference type to create (ModelReference or ExternalReference), some information is required. If a Reference cannot be created, it returns None.

        Args:
            reference_type (str): type of the reference to be created (ModelReference or ExternalReference).
            keys_dict (list): if ModelReference is selected, the required keys information in form of a JSON array.
            external_ref (str): if ExternalReference is selected, the required globally unique identifier.

        Returns:
           basyx.aas.model.Reference: BaSyx Python object of the AAS Reference.
        """
        ref_object = None
        if 'ModelReference' == reference_type:
            keys = ()
            last_type = None
            for key in keys_dict:
                basyx_key_type = await AASModelUtils.get_key_type_by_string(key['type'])
                keys += (model.Key(basyx_key_type, key['value']),)
                last_type = basyx_key_type
            ref_object = basyx.aas.model.ModelReference(
                key=keys, type_=await AASModelUtils.get_model_type_by_key_type(last_type))

        elif 'ExternalReference' == reference_type:
            ref_object = model.ExternalReference((model.Key(type_=model.KeyTypes.GLOBAL_REFERENCE,
                                                            value=external_ref),))
        return ref_object


class AASModelInfo:
    """This class contains the information related to AAS model."""
    SEMANTICID_SOFTWARE_NAMEPLATE_SUBMODEL = 'https://admin-shell.io/idta/SoftwareNameplate/1/0'
    SEMANTICID_SOFTWARE_NAMEPLATE_INSTANCE_NAME = ('https://admin-shell.io/idta/SoftwareNameplate/1/0/SoftwareNameplate'
                                              '/SoftwareNameplateInstance/InstanceName')
    SEMANTIC_ID_SOFTWARE_NAMEPLATE_CONFIG_PATHS = ('https://admin-shell.io/idta/SoftwareNameplate/1/0'
                                                   '/SoftwareNameplate/SoftwareNameplateInstance/ConfigurationPaths')
    SEMANTIC_ID_SOFTWARE_NAMEPLATE_CONFIG_TYPE = ('https://admin-shell.io/idta/SoftwareNameplate/1/0/SoftwareNameplate'
                                                  '/SoftwareNameplateInstance/ConfigurationType')
    SEMANTIC_ID_SOFTWARE_NAMEPLATE_CONFIG_URI = ('https://admin-shell.io/idta/SoftwareNameplate/1/0/SoftwareNameplate'
                                                 '/SoftwareNameplateInstance/ConfigurationURI')

    # TODO pasar aqui todos los IDs requeridos en el AAS (p.e el de AID submodel)



import asyncio
import types

import psutil

from smia.logic.services_utils import AgentServiceUtils


class AgentServices:
    """
    This class contains all the methods related to the agent services. As well as the asset has services that can be
    exposed, the agent also has services that can be exposed and used during the execution of the software.
    """

    def __init__(self, agent_object):
        """
        The constructor method adds the object of the agent to have access to its information.

        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
        """

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

        # The services dictionary contains all available services of the agent, with its associated executable methods
        self.services = {}

        # The Lock object is used to manage the access to global service dictionary
        self.lock = asyncio.Lock()

        asyncio.run(self.save_agent_service('RAM_memory_function', self.get_software_ram_memory))   # TODO no hacerlo asi, rellenarlo de otra forma (p.e. por metodos de extension)

    # ------------------------
    # Services general methods
    # ------------------------
    async def get_agent_service_by_id(self, service_id):
        """
        This method gets the agent service by its identifier. It returns None if the service is not found.

        Args:
            service_id (str): unique identifier of the agent service.

        Returns:
            method: executable method of the agent service.
        """
        async with self.lock:
            if service_id not in self.services:
                return None
            else:
                return self.services[service_id]

    async def save_agent_service(self, service_id, service_method):
        """
        This method adds a new agent service with a given identifier and the associated execution method.

        Args:
            service_id (str): unique identifier of the agent service.
            service_method: execution method associated to the agent service.
        """
        # If it is an external function, it is bounded as a method of AgentServices class. This ensures that self is
        # automatically passed when the method is called
        service_method = types.MethodType(service_method, self)
        async with self.lock:
            self.services[service_id] = service_method.__func__     # The executable function is saved

    async def execute_agent_service_by_id(self, service_id, **kwargs):
        """
        This method executes the agent service by its identifier. The parameters of the method with their values are
        available in kwargs.

        Args:
            service_id (str): identifier of the agent service.
            **kwargs: received parameters with the values.

        Returns:
            object: the result of the executed agent service ('OK' if the service does not return anything).
        """
        # First, the method of the service via its identifier is got
        service_method = await self.get_agent_service_by_id(service_id)
        if service_method is None:
            raise KeyError(f"Agent service with identifier {service_id} does not exist in this DT.")
        else:
            # Parameters adapted to the types required in the method are obtained
            adapted_params = await AgentServiceUtils.get_adapted_service_parameters(service_method, **kwargs)

            # Eventually, the method with the transformed arguments is called, and it is waited for the result
            result = await AgentServiceUtils.safe_execute_agent_service(service_method, **adapted_params)
            if result is not None:
                return result
            else:
                return "OK"

    # ----------------------
    # Agent Services methods
    # ----------------------
    async def get_software_ram_memory(self):
        """
        This agent service gets the current used RAM memory of the software.

        Returns:
            float: the current used RAM memory of the software.
        """
        return psutil.virtual_memory().percent





import logging
from spade.behaviour import State

from smia.behaviours.idle_behaviour import IdleBehaviour
from smia.utilities import smia_archive_utils

_logger = logging.getLogger(__name__)


class StateIdle(State):
    """
    This class contains the Idle state of the SMIA.
    """

    async def run(self):
        """
        This method implements the idle state of the common SMIA. Here all requests services are handled,
        both from ACL of another SMIA or from the own SMIA.
        """

        _logger.info("## STATE 4: IDLE ##")
        smia_archive_utils.update_status('Idle')

        idle_behav = IdleBehaviour(self.agent)
        self.agent.add_behaviour(idle_behav)



import logging
from spade.behaviour import CyclicBehaviour

_logger = logging.getLogger(__name__)


class IdleBehaviour(CyclicBehaviour):
    """
    This class implements the behaviour responsible for the idle state tasks.
    """

    def __init__(self, agent_object):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """
        _logger.info(str(self.agent.id) + ":     [Idle Behaviour]")
        _logger.info("         |___ Resource entering IDLE state...")

        # TODO pensar el comportamiento del SMIA en este estado (p.e. avisar del cambio de estado al Core,
        #  si llegan mensajes ACL almacenarlos para su posterior procesamiento cuando vuelva a running...)




import logging

from spade.behaviour import OneShotBehaviour

_logger = logging.getLogger(__name__)


class HandleSvcResponseBehaviour(OneShotBehaviour):
    """
    This class implements the behaviour that handles all the service responses that the SMIA has received. This
    response can arrive from an FIPA-ACL message or from the AAS Core as an interaction message. This is a
    OneShotBehaviour because it handles an individual service response and then kills itself.
    """

    # TODO PENSAR SI ELIMINAR ESTA CLASE Y AGRUPAR TANTO Requests como Responses en 'HandleSvcBehaviour', ya que para
    #  CSS solo hay CapabilityBehaviour. Dentro de esta se podria analizar la performativa (como ya se hace), para ver
    #  si es una peticion o una respuesta
    def __init__(self, agent_object, svc_resp_interaction_type, svc_resp_data):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
            svc_resp_interaction_type (str): the type of the service response interaction (:term:`Inter AAS Interaction`
            or :term:`Intra AAS Interaction`)
            svc_resp_data (dict): all the information about the service response
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object
        self.svc_resp_interaction_type = svc_resp_interaction_type
        self.svc_resp_data = svc_resp_data

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("HandleSvcResponseBehaviour starting...")

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """

        # TODO hay que pensar bien como identificar las peticiones de servicios y sus respuestas. En las interacciones
        #  Manager-Core tenemos el interactionID, con lo que es muy facil mapear la respuesta con su peticion. En
        #  cambio, si esa interaccion viene dada por una peticion anterior en ACL, hay que pensar como relacionar todas
        #  estar interacciones. Una solucion puede ser el atributo thread que propone FIP-ACL (o conversation ID). En
        #  este caso, el thread se enviara en todos los mensajes ACL, y este tambien se puede añadir en las
        #  interacciones Manager-Core. De esta forma, si una peticion de ACL exige una peticion de interaccion
        #  Manager-Core, en esta segunda se añadirá el mismo thread, de modo que podremos relacionar todas las
        #  peticiones-respuestas del mismo thread. Un servicio solo se completará del todo si todas las peticiones de
        #   ese thread tienen su respuesta

        # First, the service type of the request is obtained
        match self.svc_resp_data['serviceType']:
            case "AssetRelatedService":
                await self.handle_asset_related_service_response()
            case "AASInfrastructureServices":
                await self.handle_aas_infrastructure_svc_response()
            case "AASservices":
                await self.handle_aas_services_response()
            case "SubmodelServices":
                await self.handle_submodel_service_response()
            case _:
                _logger.error("Service type not available.")

    # ------------------------------------------
    # Methods to handle of all types of services
    # ------------------------------------------
    async def handle_asset_related_service_response(self):
        """
        This method handles Asset Related Services. These services are part of I4.0 Application Component (application
        relevant).
        """
        # TODO pensar como se gestionaria este caso. Uno que se me ocurre es que llega el ganador de una negociacion
        #  solicitada por el Core.
        _logger.info("Asset Related Service response.")
        match self.svc_resp_data['serviceID']:
            case 'negotiationResult':
                _logger.info("The result of the negotiation is due to a previous start negotiation request from "
                             "the AAS Core.")
            case 'ACLmessageResult':
                _logger.info("The Inter AAS response is due to a previous Intra AAS request from the AAS Core")
            case _:
                _logger.info("The serviceID is not one of the defaults, so the response is due to a previous "
                             "Intra AAS request from the AAS Core (the serviceID is the one that is related "
                             "with the service requested).")
                # The serviceID is updated as it is not one of the defaults
                self.svc_resp_data['serviceID'] = 'svcRequestResult'
        # TODO


    async def handle_aas_infrastructure_svc_response(self):
        """
        This method handles AAS Infrastructure Services. These services are part of I4.0 Infrastructure Services (
        Systemic relevant). They are necessary to create AASs and make them localizable and are not offered by an
        AAS, but by the platform (computational infrastructure). These include the AAS Create Service (for creating
        AASs with unique identifiers), AAS Registry Services (for registering AASs) and AAS Exposure and Discovery
        Services (for searching for AASs).

        """
        _logger.info('AAS Infrastructure Service response: ' + str(self.svc_resp_data))

    async def handle_aas_services_response(self):
        """
        This method handles AAS Services. These services serve for the management of asset-related information through
        a set of infrastructure services provided by the AAS itself. These include Submodel Registry Services (to list
        and register submodels), Meta-information Management Services (including Classification Services, to check if the
        interface complies with the specifications; Contextualization Services, to check if they belong together in a
        context to build a common function; and Restriction of Use Services, divided between access control and usage
        control) and Exposure and Discovery Services (to search for submodels or asset related services).

        """
        _logger.info('AAS Service response: ' + str(self.svc_resp_data))

    async def handle_submodel_service_response(self):
        """
        This method handles Submodel Services. These services are part of I4.0 Application Component (application
        relevant).

        """
        # TODO, en este caso tendra que comprobar que submodelo esta asociado a la peticion de servicio. Si el submodelo
        #  es propio del SMIA, podra acceder directamente y, por tanto, este behaviour sera capaz de realizar el
        #  servicio completamente. Si es un submodelo del AAS Core, tendra que solicitarselo
        _logger.info('Submodel Service response: '  + str(self.svc_resp_data))



import asyncio

import aioxmpp
from aioxmpp import PresenceState, JID
from spade.agent import Agent
import logging

from smia.aas_model.extended_aas_model import ExtendedAASModel
from smia.css_ontology.capability_skill_ontology import CapabilitySkillOntology
from smia.logic.agent_services import AgentServices
from smia.logic.exceptions import AASModelReadingError
from smia.states.state_running import StateRunning
from smia.states.state_stopping import StateStopping
from smia.utilities import properties_file_utils
from smia.utilities.general_utils import SMIAGeneralInfo
from smia.behaviours.aas_fsm_behaviour import AASFSMBehaviour
from smia.states.state_booting import StateBooting
from smia.utilities.general_utils import GeneralUtils

_logger = logging.getLogger(__name__)


class SMIAAgent(Agent):
    """
    This is the top level in the hierarchy of SPADE Agents. It extends the own class Agent of SPADE. The :term:`SMIA`
    Agent will be the generic and from which all other types of SMIAs will start.
    """

    acl_svc_requests = {}   #: Dictionary to save FIPA-ACL service requests
    acl_svc_responses = {}   #: Dictionary to save FIPA-ACL service responses
    negotiations_data = {}  #: Dictionary to save negotiations related information
    aas_model = None  #: Object with the extended AAS model
    css_ontology = None  #: Object with the Capability-Skill-Service ontology
    asset_connections = None  #: Class with the Asset Connection methods
    agent_services = None   #: Class with the all services of the Agent
    lock = None  #: Asyncio Lock object for secure access to shared SMIA objects

    # def __init__(self, jid: str = None, password: str = None, verify_security: bool = False):
    # def __init__(self, jid: str = None, password: str = None, port: int = 5222, verify_security: bool = False):     # For v4.0.0 and more
    def __init__(self, jid: str = None, password: str = None, *args, **kwargs):    # Generic constructor

        # The AAS_ID will be set in the associated ConfigMap, within the general-information of the AAS
        if jid is None:
            jid = properties_file_utils.get_dt_general_property('agentID')
        if '@' not in jid:
            # The XMPP server of the MAS will also be set in the associated ConfiMap
            xmpp_server = properties_file_utils.get_dt_general_property('xmpp-server')

            # Build the agent jid and password
            jid = jid + '@' + xmpp_server
        if password is None:
            password = properties_file_utils.get_dt_general_property('password')

        super().__init__(jid, password, *args, **kwargs)
        # super().__init__(jid, password, port, verify_security)      # For v4.0.0 and more

        # The banner of the program is printed
        GeneralUtils.print_smia_banner()

        self.initialize_smia_attributes()

    def initialize_smia_attributes(self):
        """
        This method initializes all the attributes of the SMIA.
        """
        # Objects to store the information related to ACL services are initialized
        self.acl_messages_id = 0  # It is reset
        self.acl_svc_requests = {}
        self.acl_svc_responses = {}

        # Object to store the information related to negotiations is initialized
        self.negotiations_data = {}

        # The object with the CSS ontology and useful methods is initialized
        self.css_ontology = CapabilitySkillOntology()

        # The object with the Extended AAS model and useful methods is initialized
        self.aas_model = ExtendedAASModel()

        # The object with the AssetConnection class is initialized. At this point, as an empty JSON
        if self.asset_connections is None:  # With ExtensibleSMIA asset connections can be added before
            self.asset_connections = {}

        # The class with all AgentServices is initialized
        if self.agent_services is None:  # With ExtensibleSMIA asset connections can be added before
            self.agent_services = AgentServices(self)

        # The Lock object is used to manage the access to global agent attributes (request and response dictionaries,
        # interaction id number...)
        self.lock = asyncio.Lock()

    async def setup(self):
        """
        This method performs the common setup of all types of :term:`SMIAs <SMIA>`. It defines the Finite State Machine
        (FSM) of the general SMIA Agent.
        """
        _logger.info(f"Setting up {self.jid} SMIA...")

        # First, the FSMBehaviour is instantiated
        fsm_behaviour = AASFSMBehaviour()

        # TODO HACER AHORA PENSAR SI TIENE ESTADO IDLE SMIA GENERICO (en SMIAResource esta definido con IDLE)

        # A common SMIA has three states
        fsm_behaviour.add_state(name=SMIAGeneralInfo.BOOTING_STATE_NAME, state=StateBooting(), initial=True)
        fsm_behaviour.add_state(name=SMIAGeneralInfo.RUNNING_STATE_NAME, state=StateRunning())
        fsm_behaviour.add_state(name=SMIAGeneralInfo.STOPPING_STATE_NAME, state=StateStopping())

        # Transitions are defined to determine from which state to which other state you are allowed to move to.
        fsm_behaviour.add_transition(source=SMIAGeneralInfo.BOOTING_STATE_NAME,
                                     dest=SMIAGeneralInfo.RUNNING_STATE_NAME)
        fsm_behaviour.add_transition(source=SMIAGeneralInfo.RUNNING_STATE_NAME,
                                     dest=SMIAGeneralInfo.STOPPING_STATE_NAME)

        # The FSM behaviour is added to the agent
        self.add_behaviour(fsm_behaviour)
        _logger.info(f"{self.jid} setup finished correctly.")

    # ----------------------------------------------
    # Methods related to shared objects of the agent
    # ----------------------------------------------
    async def save_new_acl_svc_request(self, thread, request_data):
        """
        This method adds a new :term:`ACL <FIPA-ACL>` Service Request to the global acl service requests dictionary of
        the SMIA.

        Args:
            thread (str): thread of the ACL Service Request.
            request_data (dict): all the information of the ACL Service Request in JSON format.
        """
        async with self.lock:  # safe access to a shared object of the agent
            self.acl_svc_requests[thread] = request_data

    async def save_acl_svc_response(self, thread, response_data):
        """
        This method adds a specific :term:`Inter SMIA interaction` response to the global responses dictionary of the
        SMIA for this type of interaction.

        Args:
            thread (str): thread of the ACL Service response.
            response_data (dict): all the information of the ACL Service response in JSON format.
        """
        async with self.lock:  # safe access to a shared object of the agent
            self.acl_svc_responses[thread] = response_data

    async def remove_acl_svc_request(self, thread):
        """
        This method removes an ACL Service Request from the global acl service requests dictionary of the SMIA.

        Args:
            thread (str): thread of the ACL Service Request.
        """
        async with self.lock:  # safe access to a shared object of the agent
            self.acl_svc_requests.pop(thread, None)

    async def get_acl_svc_request(self, thread):
        """
        This method gets the information of an ACL Service Request from the global acl service requests dictionary of
        the SMIA using the thread.
        Args:
            thread (str): thread of the ACL Service Request.

        Returns:
            dict: all information of the ACL Service Request in JSON format (null if the thread does not exist).
        """
        async with self.lock:  # safe access to a shared object of the agent
            if thread in self.acl_svc_requests:
                return self.acl_svc_requests[thread]
            else:
                return None

    async def save_negotiation_data(self, thread, neg_data):
        """
        This method saves the information of a specific negotiation in which the SMIA has participated. The data
        is stored in the global object for all negotiations of the SMIA.

        Args:
            thread (str): thread of the negotiation
            neg_data (dict): all the information of the specific negotiation
        """
        async with self.lock:  # safe access to a shared object of the agent
            self.negotiations_data[thread] = neg_data

    async def save_asset_connection_class(self, interface_reference, asset_connection):
        """
        This method adds a new asset connection to the global variable of the agent.

        Args:
            interface_reference (basyx.aas.model.ModelReference): reference of the interface of the AssetConnection
            asset_connection: class with all information about the AssetConnection
        """
        async with self.lock:  # safe access to a shared object of the agent
            self.asset_connections[interface_reference] = asset_connection

    async def get_asset_connection_class_by_ref(self, asset_connection_ref):
        """
        This method gets the asset connection class using its reference.

        Args:
            asset_connection_ref (basyx.aas.model.ModelReference): reference of the asset connection

        Returns:
            assetconnection.asset_connection: class of the asset connection
        """
        async with self.lock:  # safe access to a shared object of the agent
            for conn_ref, conn_class in self.asset_connections.items():
                if conn_ref == asset_connection_ref:
                    return conn_class
            raise AASModelReadingError("There is not asset connection class linked to {}".format(asset_connection_ref),
                                       asset_connection_ref, "MissingAssetConnectionClass")

    async def get_all_asset_connections(self):
        """
        This method returns all asset connections of the agent.

        Returns:
            dict: dictionary wil all asset connections
        """
        async with self.lock:  # safe access to a shared object of the agent
            return self.asset_connections


"""
This is the main package of SMIA, that includes all source code and all subpackages.
"""

__author__ = """Ekaitz Hurtado"""
__email__ = "ekaitz.hurtado@ehu.eus"
__version__ = '0.2.1'

__all__ = ['launchers', 'agents', 'utilities', 'assetconnection']

import logging
import ntpath

import spade

from .aas_model.aas_model_utils import AASModelUtils
from .logic.exceptions import CriticalError
from .utilities.aas_model_extension_utils import AASModelExtensionUtils
from .utilities import properties_file_utils, smia_archive_utils
from .utilities.general_utils import GeneralUtils
from .utilities.smia_general_info import SMIAGeneralInfo


def initial_self_configuration():
    """
    This method executes the initial configuration of the SMIA software.
    """
    # First, the BaSyx Python SDK is extended to have all new methods available
    AASModelExtensionUtils.extend_basyx_aas_model()

    # Initialize SMIA archive
    smia_archive_utils.initialize_smia_archive()

    # Configure logging
    GeneralUtils.configure_logging()


def run(agent_object):
    """
    This method runs the SMIA software with a given agent.
    """

    _logger = logging.getLogger(__name__)

    if agent_object is None:
        _logger.error('To launch SMIA, an agent must be passed to the method "smia.run".')
        return

    async def main():
        await agent_object.start()

        await spade.wait_until_finished(agent_object)

        # In the general properties file you can select the web interface (provided by SPADE).
        web_ui = properties_file_utils.get_dt_general_property('web-ui')
        if web_ui.lower() in ('yes', 'true', 't', '1'):
            # bool(string) cannot be used as it is true as long as the string is not empty.
            agent_object.web.start(hostname="0.0.0.0", port="10002")

    spade.run(main())


def load_aas_model(file_path):
    """
    This method loads the AAS model using a given path to the AASX package file.

    Args:
        file_path (str): path to the AASX package file.
    """
    # If the user has not run the initial self-configuration method, it is now executed
    initial_self_configuration()

    _logger = logging.getLogger(__name__)
    # TODO At the moment it only collects models in AASX, think about whether to leave option to XML and JSON as well.

    if file_path is None:
        _logger.error("The file path to the AAS model is None, so it cannot be loaded.")
        return

    # The variable with the AAS model file name is updated
    aas_model_file_name = ntpath.split(file_path)[1] or ntpath.basename(ntpath.split(file_path)[0])
    # SMIAGeneralInfo.CM_AAS_MODEL_FILENAME = aas_model_file_name
    GeneralUtils.update_aas_model(aas_model_file_name)

    # The file will be copied into the SMIA archive
    try:
        smia_archive_utils.copy_file_into_archive(file_path, SMIAGeneralInfo.CONFIGURATION_AAS_FOLDER_PATH)
    except Exception as e:
        raise CriticalError('It is not possible to copy the specified AAS model into the SMIA Archive, so the SMIA '
                      'cannot be started. Reason: {}'.format(e))
    _logger.info("AAS model {} copied to the SMIA Archive.".format(SMIAGeneralInfo.CM_AAS_MODEL_FILENAME))

    # When the AAS model is inside the SMIA archive, it will be checked if it is valid
    try:
        config_file_path = AASModelUtils.get_configuration_file_path_from_standard_submodel()
        init_config_file_name = ntpath.split(config_file_path)[1] or ntpath.basename(ntpath.split(config_file_path)[0])
        config_file_bytes = AASModelUtils.get_file_bytes_from_aasx_by_path(config_file_path)
        properties_file_utils.update_properties_file_by_bytes(config_file_bytes)
        # with open(SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH + '/' + init_config_file_name, "wb") as binary_file:
        #     binary_file.write(config_file_bytes)  # Write bytes to file
        # SMIAGeneralInfo.CM_GENERAL_PROPERTIES_FILENAME = init_config_file_name
    except Exception as e:
        _logger.warning("The AAS model does not contain the initialization configuration file. Make sure that it is "
                        "not necessary.")

        # _logger.error(e)



"""
This package includes all general utilities to be used by any kind of class. The utilities are related to the AAS
Archive, Submodel and ConfigMap.
"""

__all__ = ['aas_model_extension_utils', 'properties_file_utils', 'general_utils', 'smia_archive_utils']  # TODO Test to export only a part of the source code in the pypi package


import logging
from spade.behaviour import OneShotBehaviour

from smia.utilities import smia_archive_utils

_logger = logging.getLogger(__name__)


class InitAASarchiveBehaviour(OneShotBehaviour):
    """
    This class implements the behaviour responsible for initialize the AAS Archive, performing the necessary actions
    to let the AAS Archive in the initial conditions to start the main program.
    """

    def __init__(self, agent_object):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """
        # Create the status file
        smia_archive_utils.create_status_file()

        # Create the interaction files
        # AAS_Archive_utils.create_interaction_files()

        # Create log file
        smia_archive_utils.create_log_files()

        _logger.info("AAS Archive initialized.")



""" File to group useful methods for accessing and managing the AAS Archive."""
import json
import logging
import os
import shutil
import time
from shutil import SameFileError

from smia.utilities import properties_file_utils
from smia.utilities.fipa_acl_info import ServiceTypes
from smia.utilities.smia_general_info import SMIAGeneralInfo
from smia.utilities.general_utils import GeneralUtils

_logger = logging.getLogger(__name__)


def initialize_smia_archive():
    """
    This method initializes the SMIA Archive, performing the necessary actions to let the archive in the initial
    necessary conditions to start the software.
    """
    # Create the required folders
    create_archive_folders()

    # Create the status file
    create_status_file()

    # Create the properties file
    create_properties_file()

    # Create log file
    create_log_files()

    _logger.info("SMIA Archive initialized.")


# ------------------------
# Methods related to files
# ------------------------
def safe_open_file(file_path):
    """
    This method opens a file in a safe way. If the file does not exist, it will be created, and then opened.

    Args:
        file_path (str): The path to the file to be opened.
    """
    try:
        return open(file_path, 'x')
    except FileExistsError as e:
        return open(file_path, 'w')

def create_archive_folders():
    """
    This method ensures that all main folders of the archive exist, as well ass the parent archive folder.
    """
    required_folders_paths = [SMIAGeneralInfo.SMIA_ARCHIVE_PATH, SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH,
                              SMIAGeneralInfo.CONFIGURATION_AAS_FOLDER_PATH, SMIAGeneralInfo.STATUS_FOLDER_PATH,
                              SMIAGeneralInfo.LOG_FOLDER_PATH]
    for folder_path in required_folders_paths:
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)


def create_status_file():
    """This method creates the status file of the SMIA and sets it to "initializing". If the file exists because
    the SMIA has been restarted without terminating the Pod where it is running, the status file will be
    rewritten."""
    # The status information file will be a list of all the statuses the software has gone through.
    initial_status_info = [{'name': 'SMIA', 'status': 'Initializing',
                            'timestamp': GeneralUtils.get_current_timestamp()}]
    if not os.path.exists(SMIAGeneralInfo.STATUS_FOLDER_PATH):
        # If necessary, the status folder is created
        os.mkdir(SMIAGeneralInfo.STATUS_FOLDER_PATH)
    # try:
    #     status_file = open(SMIAGeneralInfo.STATUS_FOLDER_PATH + SMIAGeneralInfo.SMIA_STATUS_FILE_NAME, 'x')
    # except FileExistsError as e:
    #     status_file = open(SMIAGeneralInfo.STATUS_FOLDER_PATH + SMIAGeneralInfo.SMIA_STATUS_FILE_NAME, 'w')
    status_file = safe_open_file(SMIAGeneralInfo.STATUS_FOLDER_PATH + '/' + SMIAGeneralInfo.SMIA_STATUS_FILE_NAME)
    json.dump(initial_status_info, status_file)
    status_file.close()

def create_properties_file():
    """
    This method creates the properties file of the SMIA adding default values for each attribute. If the file exists
    because the SMIA has been restarted without terminating the Pod where it is running, the file will be
    rewritten.
    """
    properties_file_utils.create_empty_file()


def create_interaction_files():
    """This method creates the necessary interaction files to exchange information between AAS Core and AAS Manager."""

    # First interaction folders are created
    os.mkdir(SMIAGeneralInfo.CORE_INTERACTIONS_FOLDER_PATH)
    os.mkdir(SMIAGeneralInfo.MANAGER_INTERACTIONS_FOLDER_PATH)

    # Then the interaction files are added in each folder
    with (open(SMIAGeneralInfo.CORE_INTERACTIONS_FOLDER_PATH + SMIAGeneralInfo.SVC_REQUEST_FILE_SUBPATH,
               'x') as core_requests_file,
          open(SMIAGeneralInfo.CORE_INTERACTIONS_FOLDER_PATH + SMIAGeneralInfo.SVC_RESPONSE_FILE_SUBPATH,
               'x') as core_responses_file,
          open(SMIAGeneralInfo.MANAGER_INTERACTIONS_FOLDER_PATH + SMIAGeneralInfo.SVC_REQUEST_FILE_SUBPATH,
               'x') as manager_requests_file,
          open(SMIAGeneralInfo.MANAGER_INTERACTIONS_FOLDER_PATH + SMIAGeneralInfo.SVC_RESPONSE_FILE_SUBPATH,
               'x') as manager_responses_file):
        core_requests_file.write('{"serviceRequests": []}')
        core_requests_file.close()

        manager_requests_file.write('{"serviceRequests": []}')
        manager_requests_file.close()

        core_responses_file.write('{"serviceResponses": []}')
        core_responses_file.close()

        manager_responses_file.write('{"serviceResponses": []}')
        manager_responses_file.close()


def create_log_files():
    """This method creates the necessary log files to save services information."""

    if not os.path.exists(SMIAGeneralInfo.LOG_FOLDER_PATH):
        # If necessary, the status folder is created
        os.mkdir(SMIAGeneralInfo.LOG_FOLDER_PATH)

    all_log_folders = [SMIAGeneralInfo.SVC_LOG_FOLDER_PATH, SMIAGeneralInfo.ERROR_LOG_FOLDER_PATH]

    # Then the log files are added in each folder
    all_svc_log_file_names = [SMIAGeneralInfo.ASSET_RELATED_SVC_LOG_FILENAME,
                              SMIAGeneralInfo.AAS_INFRASTRUCTURE_SVC_LOG_FILENAME,
                              SMIAGeneralInfo.AAS_SERVICES_LOG_FILENAME, SMIAGeneralInfo.SUBMODEL_SERVICES_LOG_FILENAME,
                              SMIAGeneralInfo.SUBMODEL_CSS_LOG_FILENAME]
    for log_folder in all_log_folders:
        if not os.path.exists(log_folder):
            # If necessary, folders are also created for each log
            os.mkdir(log_folder)
        for log_file_name in all_svc_log_file_names:
            with safe_open_file(log_folder + '/' + log_file_name) as log_file:
                log_file.write('[]')
                log_file.close()


def save_cli_added_files(init_config=None, aas_model=None):
    """
    This method saves the specified files in the Command Line Interface into the SMIA Archive.

    Args:
        init_config (str): path to the initialization configuration properties file.
        aas_model (str): path to the AAS model file.
    """
    # for cli_file in [init_config, aas_model]:
    #     if cli_file is not None:
    #         copy_file_into_archive(cli_file, SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH)
    # The initialization configuration properties file is stored in the main configuration folder of the SMIA Archive
    if init_config is not None:
        copy_file_into_archive(init_config, SMIAGeneralInfo.CONFIGURATION_FOLDER_PATH)
    # The AAS model file is stored in the AAS-related configuration folder of the SMIA Archive
    if aas_model is not None:
        copy_file_into_archive(aas_model, SMIAGeneralInfo.CONFIGURATION_AAS_FOLDER_PATH)


def copy_file_into_archive(source_file, dest_file):
    """
    This method copies a file into the SMIA archive.

    Args:
        source_file (str): path of the source file.
        dest_file (str): path of the destiny file (it must be inside the SMIA archive).
    """
    # If the archive does not exist, it must be initialized
    if not os.path.exists(SMIAGeneralInfo.SMIA_ARCHIVE_PATH):
        create_archive_folders()

    if SMIAGeneralInfo.SMIA_ARCHIVE_PATH not in dest_file:
        dest_file = SMIAGeneralInfo.SMIA_ARCHIVE_PATH + '/' + dest_file
    try:
        shutil.copy(source_file, dest_file)
    except SameFileError as e:
        _logger.info("The {} file is already in SMIA archive.".format(source_file))


# ----------------------
# Methods related to log
# ----------------------
def get_log_file_by_service_type(log_type, svc_type):
    """
    This method obtains the path to the log file associated to the type of the service.
    Args:
        log_type (str): the type of the log.
        svc_type (str): type of the service.

    Returns:
        str: path to the log file
    """
    log_file_path = None
    if log_type == 'services':
        log_folder_path = SMIAGeneralInfo.SVC_LOG_FOLDER_PATH
    elif log_type == 'errors':
        log_folder_path = SMIAGeneralInfo.ERROR_LOG_FOLDER_PATH
    else:
        _logger.error("Log type not available.")
        return None

    match svc_type:
        case ServiceTypes.ASSET_RELATED_SERVICE:
            log_file_path = log_folder_path + '/' + SMIAGeneralInfo.ASSET_RELATED_SVC_LOG_FILENAME
        case ServiceTypes.AAS_INFRASTRUCTURE_SERVICE:
            log_file_path = log_folder_path + '/' + SMIAGeneralInfo.AAS_INFRASTRUCTURE_SVC_LOG_FILENAME
        case ServiceTypes.AAS_SERVICE:
            log_file_path = log_folder_path + '/' + SMIAGeneralInfo.AAS_SERVICES_LOG_FILENAME
        case ServiceTypes.SUBMODEL_SERVICE:
            log_file_path = log_folder_path + '/' + SMIAGeneralInfo.SUBMODEL_SERVICES_LOG_FILENAME
        case ServiceTypes.CSS_RELATED_SERVICE:
            log_file_path = log_folder_path + '/' + SMIAGeneralInfo.SUBMODEL_CSS_LOG_FILENAME
        case _:
            _logger.error("Service type not available.")
    return log_file_path


def save_completed_svc_log_info(started_timestamp, finished_timestamp, acl_info, result, svc_type):
    """
    This method saves the information about a realized service in the log file associated to the type of the service.

    Args:
        started_timestamp (int): the timestamp when the service has been requested.
        finished_timestamp (int): the timestamp when the service has finished.
        acl_info (dict): JSON object with the information of the service request.
        result: the result of the service execution.
        svc_type (str): type of the service.
    """
    # First, the required paths are obtained
    log_file_path = get_log_file_by_service_type('services', svc_type)

    # Then, the information object is built and added
    svc_log_info = {
        'startedTimestamp': started_timestamp,
        'finishedTimestamp': finished_timestamp,
        'status': 'Completed',
        'requestInfo': acl_info,
        'result': result
    }
    log_file_json = file_to_json(log_file_path)
    log_file_json.append(svc_log_info)

    # Finally, the file is updated
    update_json_file(log_file_path, log_file_json)

    _logger.info("Saved new completed service information in the log.")

def get_file_by_extension(folder_path, file_extension):
    """
    This method obtains the path of a file inside a given folder with a given extension.

    Args:
        folder_path (str): path to the folder to be searched.
        file_extension (str): extension of the file (without dot '.').

    Returns:
        str: path of the file to search (None if it does not exist).
    """
    if not os.path.exists(folder_path):
        _logger.warning('The folder path to be searched does not exist')
        return None
    dir_files = os.listdir(folder_path)
    for file in dir_files:
        file_name, file_ext = os.path.splitext(file)
        if file_extension == file_ext:
            return  folder_path + '/' + file
    _logger.warning('A file with extension {} was not found inside the {} folder.'.format(file_extension, folder_path))
    return None

def save_svc_error_log_info(occurrence_timestamp, acl_info, reason, svc_type):
    """
    This method saves the information about a realized service in the log file associated to the type of the service.

    Args:
        occurrence_timestamp (int): the timestamp when the error occurred.
        acl_info (dict): JSON object with the information of the service request.
        reason (str): the reason of the error.
        svc_type (str): type of the service.
    """
    # First, the required paths are obtained
    log_file_path = get_log_file_by_service_type('errors', svc_type)

    # Then, the information object is built and added
    svc_log_info = {
        'timestamp': occurrence_timestamp,
        'status': 'Error',
        'requestInfo': acl_info,
        'reason': reason
    }
    log_file_json = file_to_json(log_file_path)
    log_file_json.append(svc_log_info)

    # Finally, the file is updated
    update_json_file(log_file_path, log_file_json)

    _logger.info("Saved new completed service information in the log.")


# -------------------------
# Methods related to status
# -------------------------
def update_status(new_status):
    """
    This method updated the status of an SMIA instance.

    Args:
        new_status (str): the new status of the SMIA instance.
    """
    status_file_path = SMIAGeneralInfo.STATUS_FOLDER_PATH + '/' + SMIAGeneralInfo.SMIA_STATUS_FILE_NAME
    status_file_json = file_to_json(status_file_path)
    status_file_json.append({'name': 'SMIA', 'status': new_status,
                             'timestamp': GeneralUtils.get_current_timestamp()})
    update_json_file(status_file_path, status_file_json)


def get_status(entity):
    """
    This method gets the status of the requested entity.

    Args:
        entity (str): The entity to get the status for.

    Returns:
        dict: status in JSON format.
    """
    status_file_json = None
    if entity == "Manager":
        status_file_json = file_to_json(SMIAGeneralInfo.SMIA_STATUS_FILE_PATH)
    elif entity == "Core":
        status_file_json = file_to_json(SMIAGeneralInfo.CORE_STATUS_FILE_PATH)
    return status_file_json['status']


def check_core_initialization():
    """This method checks if the core has initialized so the Manager can be started."""
    while True:
        if os.path.isfile(SMIAGeneralInfo.CORE_STATUS_FILE_PATH) is True:
            if file_to_json(SMIAGeneralInfo.CORE_STATUS_FILE_PATH)['status'] != "Initializing":
                break
        time.sleep(1)  # waits 1s
    _logger.info('AAS Core has initialized, so the SMIA is starting.')


# ------------------------
# Methods related to JSON
# ------------------------
def file_to_json(file_path):
    """
    This method gets the content of a JSON file.

    Args:
        file_path (str): the path of the JSON file.

    Returns:
        dict: content of the file in JSON format."""
    f = open(file_path)
    try:
        content = json.load(f)
        f.close()
    except json.JSONDecodeError as e:
        _logger.error("Invalid JSON syntax:" + str(e))
        return None
    return content


def update_json_file(file_path, content):
    """
    This method updates the content of a JSON file.

    Args:
        file_path (str): the path to the JSON file.
        content (dict): the content of the JSON file.
    """
    with open(file_path, "w") as outfile:
        json.dump(content, outfile)


# ------------------------
# Methods related to XML
# ------------------------
def xml_to_file(file_path, xml_content):
    """
    This method writes the content of an XML in a file.

    Args:
        file_path (str): the path to the XML file.
        xml_content (bytes): the content of the XML file.
    """
    with open(file_path, 'wb') as xml_file:
        xml_file.write(xml_content)



"""
This is the SMIA launch file, which executes the program logic from the command and gets the AAS model by the provided
Python methods.
"""

import logging

import smia
from smia.agents.extensible_smia_agent import ExtensibleSMIAAgent
from smia.agents.smia_agent import SMIAAgent
from smia.assetconnection.asset_connection import AssetConnection

_logger = logging.getLogger(__name__)


def nuevo_svc():
    print("dsfsdffsd")

async def nuevo_svc_params(param1, param2):
    print("{}, {}".format(param1, param2))



def main():
    # First, the initial configuration must be executed
    smia.initial_self_configuration()
    _logger.info("Initializing SMIA software...")

    # Then, the AASX model is added
    # smia.load_aas_model('../examples/SMIA_Operator_article.aasx')
    smia.load_aas_model('../examples/SMIA_TransportRobot_article_1.aasx')
    # smia.load_aas_model('../examples/SMIA_TransportRobot_without_OWL.aasx')
    # smia.load_aas_model('../examples/SMIA_tutorial_1.aasx')
    # smia.load_aas_model('../smia_archive/config/SMIA_TransportRobot_with_OWL.aasx')

    # Create the agent object
    smia_agent = SMIAAgent("gcis2@xmpp.jp", "gcis1234")
    # smia_agent = SMIAAgent()

    # smia_agent = ExtensibleSMIAAgent()
    # smia_agent.add_new_agent_service('newSvc', nuevo_svc_params)

    smia.run(smia_agent)


if __name__ == '__main__':

    # Run main program with SMIA
    main()



import logging

from smia.agents.smia_agent import SMIAAgent
from smia.states.state_booting_resource import StateBootingResource
from smia.states.state_running import StateRunning
from smia.states.state_stopping import StateStopping
from smia.states.state_idle import StateIdle
from smia.utilities.general_utils import SMIAGeneralInfo
from smia.behaviours.aas_fsm_behaviour import AASFSMBehaviour

_logger = logging.getLogger(__name__)


class SMIAResourceAgent(SMIAAgent):
    """
    This is the SMIA Agent for physical resource assets. It extends the generic class AASManagerAgent.
    """

    async def setup(self):
        """
        This method performs the setup of physical resource type of Managers. It defines the Finite State Machine (FSM)
        of the SMIA Agent.
        """

        # First, the FSMBehaviour is instantiated
        fsm_behaviour = AASFSMBehaviour()

        # A common SMIA has three states
        fsm_behaviour.add_state(name=SMIAGeneralInfo.BOOTING_STATE_NAME, state=StateBootingResource(), initial=True)
        fsm_behaviour.add_state(name=SMIAGeneralInfo.RUNNING_STATE_NAME, state=StateRunning())
        fsm_behaviour.add_state(name=SMIAGeneralInfo.STOPPING_STATE_NAME, state=StateStopping())
        fsm_behaviour.add_state(name=SMIAGeneralInfo.IDLE_STATE_NAME, state=StateIdle())

        # Transitions are defined to determine from which state to which other state you are allowed to move to.
        fsm_behaviour.add_transition(source=SMIAGeneralInfo.BOOTING_STATE_NAME, dest=SMIAGeneralInfo.RUNNING_STATE_NAME)
        fsm_behaviour.add_transition(source=SMIAGeneralInfo.RUNNING_STATE_NAME, dest=SMIAGeneralInfo.STOPPING_STATE_NAME)
        fsm_behaviour.add_transition(source=SMIAGeneralInfo.RUNNING_STATE_NAME, dest=SMIAGeneralInfo.IDLE_STATE_NAME)
        fsm_behaviour.add_transition(source=SMIAGeneralInfo.IDLE_STATE_NAME, dest=SMIAGeneralInfo.RUNNING_STATE_NAME)
        fsm_behaviour.add_transition(source=SMIAGeneralInfo.IDLE_STATE_NAME, dest=SMIAGeneralInfo.STOPPING_STATE_NAME)

        # The FSM behaviour is added to the agent
        self.add_behaviour(fsm_behaviour)
        _logger.info(f"{self.jid} setup finished correctly.")



import json
import logging

from spade.behaviour import OneShotBehaviour

from smia import GeneralUtils
from smia.aas_model.aas_model_utils import AASModelUtils
from smia.logic import inter_aas_interactions_utils
from smia.logic.exceptions import RequestDataError, ServiceRequestExecutionError, AASModelReadingError, \
    AssetConnectionError
from smia.utilities import smia_archive_utils
from smia.utilities.fipa_acl_info import FIPAACLInfo, ACLJSONSchemas, ServiceTypes
from smia.utilities.smia_info import AssetInterfacesInfo

_logger = logging.getLogger(__name__)


class HandleSvcRequestBehaviour(OneShotBehaviour):
    """
    This class implements the behaviour that handles all the service requests that the SMIA has received. This
    request can arrive from an FIPA-ACL message as a :term:`Inter AAS Interaction` or from the AAS Core as an
    :term:`Intra AAS Interaction` message. This is a OneShotBehaviour because it handles an individual service request
    and then kills itself.
    """

    # TODO PENSAR SI AGRUPAR EN ESTA CLASE TANTO Requests como Responses ('HandleSvcBehaviour'), ya que para CSS solo
    #  hay CapabilityBehaviour. Dentro de esta se podria analizar la performativa (como ya se hace), para ver si es una
    #  peticion o una respuesta

    def __init__(self, agent_object, svc_req_data):
        """
        The constructor method is rewritten to add the object of the agent.

        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
            svc_req_data (dict): all the information about the service request
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object
        self.svc_req_data = svc_req_data

        self.requested_timestamp = GeneralUtils.get_current_timestamp()

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("HandleSvcRequestBehaviour starting...")

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """

        # First, the performative of the request is obtained
        # TODO NUEVO ENFOQUE (usando performative)
        match self.svc_req_data['performative']:
            case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST:  # TODO actualizar dentro de todo el codigo los usos de performativas y ontologias de FIPA-ACL
                await self.handle_request()
            case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_QUERY_IF:
                await self.handle_query_if()
            # TODO PensarOtros
            case _:
                _logger.error("Performative not available for service management.")

    # ------------------------------------------
    # Methods to handle of all types of services
    # ------------------------------------------
    async def handle_request(self):
        """
        This method handle capability requests to the DT.
        """
        # The type is analyzed to perform the appropriate service
        match self.svc_req_data['serviceType']:
            case ServiceTypes.ASSET_RELATED_SERVICE:
                # await self.handle_asset_related_svc()   # TODO
                await self.handle_asset_related_service_request()
            case ServiceTypes.AAS_INFRASTRUCTURE_SERVICE:
                await self.handle_aas_infrastructure_svc_request()  # TODO
            case ServiceTypes.AAS_SERVICE:
                await self.handle_aas_services_request()  # TODO
            case ServiceTypes.SUBMODEL_SERVICE:
                await self.handle_submodel_service_request()
            case _:
                _logger.error("Service type not available.")

    async def handle_query_if(self):
        """This method handle Query-If service requests. This request is received when the DT is asked about information
         related to a service."""
        pass
        # TODO FALTA POR HACER

    # -----------------------------------
    # Methods to handle specific services
    # -----------------------------------
    async def handle_asset_related_service_request(self):
        """
        This method handles an Asset Related Service request. These services are part of I4.0 Application Component
        (application relevant).
        """
        # TODO In this case, AssetRelatedService and AssetService are considered equivalent.
        # The asset related service will be executed using a ModelReference to the related SubmodelElement within the
        # AssetInterfacesDefinition submodel. This information is in serviceParams TODO (de momento en serviceParams, ya veremos mas adelante)
        try:
            # First, the received data is checked and validated
            await inter_aas_interactions_utils.check_received_request_data_structure(
                self.svc_req_data, ACLJSONSchemas.JSON_SCHEMA_ASSET_SERVICE_REQUEST)
            service_params = self.svc_req_data['serviceData']['serviceParams']

            # If the received data is valid, the AAS Reference object need to be created
            aas_asset_service_ref = await AASModelUtils.create_aas_reference_object(
                'ModelReference', service_params['ModelReference']['keys'])
            aas_asset_service_elem = await self.myagent.aas_model.get_object_by_reference(aas_asset_service_ref)

            # The asset connection class is also required to execute the asset related service. It is obtained from the
            # AAS asset service Reference object
            aas_asset_interface_ref = aas_asset_service_elem.get_parent_ref_by_semantic_id(
                AssetInterfacesInfo.SEMANTICID_INTERFACE)
            if aas_asset_service_elem is None:
                raise ServiceRequestExecutionError(self.svc_req_data['thread'], "The added ModelReference is not "
                                                                                " inside the AssetInterfacesDescription "
                                                                                "submodel, so it is not an AssetService.",
                                                   self.svc_req_data['serviceType'], self)
            asset_connection_class = await self.myagent.get_asset_connection_class_by_ref(aas_asset_interface_ref)

            # With all necessary information obtained, the asset related service can be executed
            _logger.assetinfo("Executing skill of the capability through an asset service...")
            received_input_data = None
            if 'serviceParameterValues' in service_params:
                received_input_data = service_params['serviceParameterValues']
            asset_service_execution_result = await asset_connection_class.execute_asset_service(
                interaction_metadata=aas_asset_service_elem,
                service_input_data=received_input_data)
            _logger.assetinfo("Skill of the capability successfully executed.")

            # The result will be sent to the requester
            await self.send_response_msg_to_sender(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM,
                                                   {'result': asset_service_execution_result})
            _logger.info("Management of the service with thread {} finished.".format(self.svc_req_data['thread']))

            # The information will be stored in the log
            smia_archive_utils.save_completed_svc_log_info(self.requested_timestamp,
                                                           GeneralUtils.get_current_timestamp(),
                                                           self.svc_req_data, str(asset_service_execution_result),
                                                           self.svc_req_data['serviceType'])

        except (RequestDataError, ServiceRequestExecutionError,
                AASModelReadingError, AssetConnectionError) as svc_request_error:
            if isinstance(svc_request_error, RequestDataError):
                svc_request_error = ServiceRequestExecutionError(self.svc_req_data['thread'],
                                                                 svc_request_error.message,
                                                                 self.svc_req_data['serviceType'], self)
            if isinstance(svc_request_error, AASModelReadingError):
                svc_request_error = ServiceRequestExecutionError(self.svc_req_data['thread'], "{}. Reason: "
                                                                                              "{}".format(
                    svc_request_error.message,
                    svc_request_error.reason), self.svc_req_data['serviceType'], self)
            if isinstance(svc_request_error, AssetConnectionError):
                svc_request_error = ServiceRequestExecutionError(self.svc_req_data['thread'],
                                                                 f"The error [{svc_request_error.error_type}] has appeared during the asset "
                                                                 f"connection. Reason: {svc_request_error.reason}.",
                                                                 self.svc_req_data['serviceType'], self)
            await svc_request_error.handle_service_execution_error()
            return  # killing a behaviour does not cancel its current run loop

    async def handle_aas_services_request(self):
        """
        This method handles AAS Services. These services serve for the management of asset-related information through
        a set of infrastructure services provided by the AAS itself. These include Submodel Registry Services (to list
        and register submodels), Meta-information Management Services (including Classification Services, to check if the
        interface complies with the specifications; Contextualization Services, to check if they belong together in a
        context to build a common function; and Restriction of Use Services, divided between access control and usage
        control) and Exposure and Discovery Services (to search for submodels or asset related services).

        """
        _logger.info('AAS Service request: ' + str(self.svc_req_data))

    async def handle_aas_infrastructure_svc_request(self):
        """
        This method handles AAS Infrastructure Services. These services are part of I4.0 Infrastructure Services
        (Systemic relevant). They are necessary to create AASs and make them localizable and are not offered by an AAS, but
        by the platform (computational infrastructure). These include the AAS Create Service (for creating AASs with unique
        identifiers), AAS Registry Services (for registering AASs) and AAS Exposure and Discovery Services (for searching
        for AASs).

        """
        _logger.info('AAS Infrastructure Service request: ' + str(self.svc_req_data))

    async def handle_submodel_service_request(self):
        """
        This method handles a Submodel Service request. These services are part of I4.0 Application Component
        (application relevant).
        """
        # The submodel service will be executed using a ModelReference or an ExternalReference, depending on the
        # requested element. This information is in serviceParams TODO (de momento en serviceParams, ya veremos mas adelante)
        try:
            # First, the received data is checked and validated
            await inter_aas_interactions_utils.check_received_request_data_structure(
                self.svc_req_data, ACLJSONSchemas.JSON_SCHEMA_SUBMODEL_SERVICE_REQUEST)

            # If the data is valid, the SubmodelElement is obtained from the AAS model. For this purpose, the
            # appropriate BaSyx object must be created
            ref_object = None
            if 'ModelReference' in self.svc_req_data['serviceData']['serviceParams']:
                ref_object = await AASModelUtils.create_aas_reference_object(
                    'ModelReference',
                    keys_dict=self.svc_req_data['serviceData']['serviceParams']['ModelReference']['keys'])
            elif 'ExternalReference' in self.svc_req_data['serviceData']['serviceParams']:
                ref_object = await AASModelUtils.create_aas_reference_object(
                    'ExternalReference',
                    external_ref=self.svc_req_data['serviceData']['serviceParams']['ExternalReference'])

            # When the appropriate Reference object is created, the requested SubmodelElement can be obtained
            requested_sme = await self.myagent.aas_model.get_object_by_reference(ref_object)

            # When the AAS object has been obtained, the request is answered
            # TODO de momento simplemente lo devolvemos en string (para mas adelante pensar si desarrollar un metodo que
            #  devuelva toda la informacion)
            sme_info = str(requested_sme)
            await self.send_response_msg_to_sender(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM,
                                                   {'requested_object': sme_info})
            _logger.info("Management of the service with thread {} finished.".format(self.svc_req_data['thread']))

            # The information will be stored in the log
            smia_archive_utils.save_completed_svc_log_info(self.requested_timestamp,
                                                           GeneralUtils.get_current_timestamp(),
                                                           self.svc_req_data, str(sme_info),
                                                           self.svc_req_data['serviceType'])

        except (RequestDataError, ServiceRequestExecutionError, AASModelReadingError) as svc_request_error:
            if isinstance(svc_request_error, RequestDataError):
                svc_request_error = ServiceRequestExecutionError(self.svc_req_data['thread'],
                                                                 svc_request_error.message,
                                                                 self.svc_req_data['serviceType'], self)
            if isinstance(svc_request_error, AASModelReadingError):
                svc_request_error = ServiceRequestExecutionError(self.svc_req_data['thread'],
                                                                 "{}. Reason: {}".format(svc_request_error.message,
                                                                                         svc_request_error.reason),
                                                                 self.svc_req_data['serviceType'], self)
            await svc_request_error.handle_service_execution_error()
            return  # killing a behaviour does not cancel its current run loop

    async def send_response_msg_to_sender(self, performative, service_params):
        """
        This method creates and sends a FIPA-ACL message with the given serviceParams and performative.

        Args:
            performative (str): performative according to FIPA-ACL standard.
            service_params (dict): JSON with the serviceParams to be sent in the message.
        """
        acl_msg = inter_aas_interactions_utils.create_inter_smia_response_msg(
            receiver=self.svc_req_data['sender'],
            thread=self.svc_req_data['thread'],
            performative=performative,
            ontology='SvcResponse',
            service_id=self.svc_req_data['serviceID'],
            service_type=self.svc_req_data['serviceType'],
            service_params=json.dumps(service_params)
        )
        await self.send(acl_msg)




from smia.utilities.fipa_acl_info import FIPAACLInfo
from smia.utilities.general_utils import GeneralUtils


class SMIAInteractionInfo:
    """This class contains the information about the SMIA interactions."""

    # Object of the standard template for service requests through ACL messages
    # -------------------------------------------------------------------------
    # TODO finalizar con el plantilla estandar final decidida para la comunicacion entre agentes
    SVC_STANDARD_ACL_TEMPLATE_CFP = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_CFP,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_REQUEST)
    SVC_STANDARD_ACL_TEMPLATE_INFORM = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_REQUEST)
    SVC_STANDARD_ACL_TEMPLATE_REQUEST = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_REQUEST)
    SVC_STANDARD_ACL_TEMPLATE_QUERY_IF = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_QUERY_IF,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_REQUEST)
    # TODO OJO! De momento añado las plantillas de capacidades tambien, pero la idea es hacer dos behaviour diferentes
    CAP_STANDARD_ACL_TEMPLATE_REQUEST = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_CAPABILITY_REQUEST)
    CAP_STANDARD_ACL_TEMPLATE_QUERY_IF = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_QUERY_IF,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_CAPABILITY_CHECKING)
    # The template for the service requests is the combination of the different possibilities
    SVC_STANDARD_ACL_TEMPLATE = (SVC_STANDARD_ACL_TEMPLATE_CFP | SVC_STANDARD_ACL_TEMPLATE_INFORM
                                 | SVC_STANDARD_ACL_TEMPLATE_REQUEST | SVC_STANDARD_ACL_TEMPLATE_QUERY_IF
                                 | CAP_STANDARD_ACL_TEMPLATE_REQUEST | CAP_STANDARD_ACL_TEMPLATE_QUERY_IF  # TODO OJO, estos son los dos las capacidades
                                 )

    # Object of the standard template for service requests through ACL messages
    # -------------------------------------------------------------------------
    NEG_STANDARD_ACL_TEMPLATE_CFP = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_CFP,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION)
    NEG_STANDARD_ACL_TEMPLATE_PROPOSE = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_PROPOSE,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION)
    NEG_STANDARD_ACL_TEMPLATE_FAILURE = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_FAILURE,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION)
    NEG_STANDARD_ACL_TEMPLATE_INFORM = GeneralUtils.create_acl_template(
        performative=FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM,
        ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION)
    # The template for the negotiations is the combination of the different possibilities
    NEG_STANDARD_ACL_TEMPLATE = (NEG_STANDARD_ACL_TEMPLATE_CFP | NEG_STANDARD_ACL_TEMPLATE_FAILURE |
                                 NEG_STANDARD_ACL_TEMPLATE_INFORM)


class AssetInterfacesInfo:
    """This class contains the information related to Asset Interfaces submodel (AssetInterfacesDescription submodel of
    IDTA)."""

    # SemanticIDs of Asset Interfaces
    SEMANTICID_INTERFACES_SUBMODEL = 'https://admin-shell.io/idta/AssetInterfacesDescription/1/0/Submodel'
    SEMANTICID_INTERFACE = 'https://admin-shell.io/idta/AssetInterfacesDescription/1/0/Interface'
    SEMANTICID_ENDPOINT_METADATA = 'https://admin-shell.io/idta/AssetInterfacesDescription/1/0/EndpointMetadata'
    SEMANTICID_INTERACTION_METADATA = 'https://admin-shell.io/idta/AssetInterfacesDescription/1/0/InteractionMetadata'
    SEMANTICID_VALUE_SEMANTICS = 'https://admin-shell.io/idta/AssetInterfacesDescription/1/0/valueSemantics'

    # SemanticIDs of W3 2011
    SUPPL_SEMANTICID_HTTP = 'http://www.w3.org/2011/http'
    SEMANTICID_INTERFACE_INTERACTION_TYPE = 'https://www.w3.org/1999/02/22-rdf-syntax-ns#type'

    # SemanticIDs of Web of Things (WoT) ontology
    SEMANTICID_INTERFACE_TITLE = 'https://www.w3.org/2019/wot/td#title'
    SEMANTICID_INTERFACE_BASE = 'https://www.w3.org/2019/wot/td#baseURI'
    SEMANTICID_INTERFACE_CONTENT_TYPE = 'https://www.w3.org/2019/wot/hypermedia#forContentType'

    SEMANTICID_INTERFACE_SECURITY_DEFINITIONS = 'https://www.w3.org/2019/wot/td#definesSecurityScheme'
    SEMANTICID_INTERFACE_NO_SECURITY_SCHEME = 'https://www.w3.org/2019/wot/security#NoSecurityScheme'

    SEMANTICID_INTERFACE_PROPERTY = 'https://www.w3.org/2019/wot/td#PropertyAffordance'
    SEMANTICID_INTERFACE_ACTION = 'https://www.w3.org/2019/wot/td#ActionAffordance'

    SEMANTICID_INTERFACE_FORMS = 'https://www.w3.org/2019/wot/td#hasForm'
    SEMANTICID_INTERFACE_HREF = 'https://www.w3.org/2019/wot/hypermedia#hasTarget'

    # TODO NUEVO
    SEMANTICID_INTERFACE_INTERACTION_DATA_QUERY = 'https://admin-shell.io/idta/AssetInterfacesDescription/1/0/dataQuery'  # TODO CUIDADO, ESTE ES NUEVO, PENSAR COMO SERIA. Se ha pensado asociar este ID con un apartado de la interfaz donde se añadirá el query para extraer la informacion de la respuesta del activo (esta query debera ir en relacion con el tipo de contenido, p.e. JSONata o JSONPath para respuestas JSON)



import json
import logging

from spade.behaviour import CyclicBehaviour

from smia import GeneralUtils
from smia.css_ontology.css_ontology_utils import CapabilitySkillOntologyUtils
from smia.logic import negotiation_utils, inter_aas_interactions_utils
from smia.logic.exceptions import CapabilityRequestExecutionError, AssetConnectionError
from smia.utilities import smia_archive_utils
from smia.utilities.fipa_acl_info import FIPAACLInfo, ServiceTypes
from smia.utilities.smia_info import AssetInterfacesInfo

_logger = logging.getLogger(__name__)


class HandleNegotiationBehaviour(CyclicBehaviour):
    """
    This class implements the behaviour that handle a particular negotiation.
    """
    myagent = None  #: the SPADE agent object of the SMIA agent.
    thread = None  #: thread of the negotiation
    neg_requester_jid = None  #: JID of the SPADE agent that has requested the negotiation
    targets = None  #: targets of the negotiation
    neg_criteria = None  #: criteria of the negotiation
    neg_value = None  #: value of the negotiation
    targets_processed = set()  #: targets that their values have been processed
    neg_value_event = None

    def __init__(self, agent_object, negotiation_info, neg_req_data):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
            negotiation_info (dict): all the required information to perform the negotiation.
            neg_req_data (dict): all the information related to the FIPA-ACL negotiation request.
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

        self.thread = negotiation_info['thread']
        self.neg_requester_jid = negotiation_info['neg_requester_jid']
        self.targets = negotiation_info['targets']
        self.neg_criteria = negotiation_info['neg_criteria']
        self.targets_processed = set()

        self.myagent.tie_break = True   # In case of equal value neg is set as tie-breaker TODO check these cases (which need to be tie-breaker?)

        self.svc_req_data = neg_req_data

        self.requested_timestamp = GeneralUtils.get_current_timestamp()

        # This event object will allow waiting for the negotiation value if it is necessary to request it from an
        # external entity (such as the AAS Core)
        # self.neg_value_event = asyncio.Event()

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("HandleNegotiationBehaviour starting...")

        try:
            #  The value of the criterion must be obtained just before starting to manage the negotiation, so that at the
            #  time of sending the PROPOSE and receiving that of the others it will be the same value. Therefore, if to
            #  obtain the value you have to make an Intra AAS interaction request, the behaviour will not be able to start
            #  managing the negotiation until you get the answer to that request (together with the requested value).
            await self.get_neg_value_with_criteria()


            # Once the negotiation value is reached, the negotiation management can begin. The first step is to send the
            # PROPOSE message with your own value to the other participants in the negotiation.
            acl_propose_msg = negotiation_utils.create_neg_propose_msg(thread=self.thread,
                                                                       targets=self.targets,
                                                                       neg_requester_jid=self.neg_requester_jid,
                                                                       neg_criteria=self.neg_criteria,
                                                                       neg_value=str(self.neg_value))
            # This PROPOSE FIPA-ACL message is sent to all participants of the negotiation (except for this SMIA)
            for jid_target in self.targets.split(','):
                if jid_target != str(self.agent.jid):
                    acl_propose_msg.to = jid_target
                    await self.send(acl_propose_msg)
                    _logger.aclinfo("ACL PROPOSE negotiation message sent to " + jid_target +
                                    " on negotiation with thread [" + self.thread + "]")
        except (CapabilityRequestExecutionError, AssetConnectionError) as cap_neg_error:
            if isinstance(cap_neg_error, AssetConnectionError):
                cap_neg_error = CapabilityRequestExecutionError('Negotiation', f"The error "
                                                                f"[{cap_neg_error.error_type}] has appeared during the"
                                                                f" asset connection. Reason: {cap_neg_error.reason}.", self)

            await cap_neg_error.handle_capability_execution_error()
            return  # killing a behaviour does not cancel its current run loop


    async def run(self):
        """
        This method implements the logic of the behaviour.
        """

        # Wait for a message with the standard ACL template for negotiating to arrive.
        msg = await self.receive(
            timeout=10)  # Timeout set to 10 seconds so as not to continuously execute the behavior.
        if msg:
            # An ACL message has been received by the agent
            _logger.aclinfo("         + PROPOSE Message received on SMIA Agent (HandleNegotiationBehaviour "
                            "in charge of the negotiation with thread [" + self.thread + "])")
            _logger.aclinfo("                 |___ Message received with content: {}".format(msg.body))

            # The msg body will be parsed to a JSON object
            msg_json_body = json.loads(msg.body)

            # The negotiation information is obtained from the message
            # criteria = msg_json_body['serviceData']['serviceParams']['criteria']
            sender_agent_neg_value = msg_json_body['serviceData']['serviceParams']['neg_value']

            # The value of this SMIA and the received value are compared
            if float(sender_agent_neg_value) > self.neg_value:
                # As the received value is higher than this SMIA value, it must exit the negotiation.
                await self.exit_negotiation(is_winner=False)
                return  # killing a behaviour does not cancel its current run loop
            if (float(sender_agent_neg_value) == self.neg_value) and not self.agent.tie_break:
                # In this case the negotiation is tied but this SMIA is not the tie breaker.
                await self.exit_negotiation(is_winner=False)
                return  # killing a behaviour does not cancel its current run loop
            # The target is added as processed in the local object (as it is a Python 'set' object there is no problem
            # of duplicate agents)
            self.targets_processed.add(str(msg.sender))
            if len(self.targets_processed) == len(self.targets.split(',')) - 1:
                # In this case all the values have already been received, so the value of this SMIA is the best
                _logger.info("The AAS has won the negotiation with thread [" + msg.thread + "]")

                # As the winner, it will reply to the sender with the result of the negotiation
                await self.send_response_msg_to_sender(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM,
                                                       {'winner': str(self.myagent.jid)})

                _logger.aclinfo("ACL response sent for the result of the negotiation request with thread ["
                                + msg.thread + "]")

                # The negotiation can be terminated, in this case being the winner
                await self.exit_negotiation(is_winner=True)

        else:
            _logger.info("         - No message received within 10 seconds on SMIA Agent (NegotiatingBehaviour)")

    async def get_neg_value_with_criteria(self):
        """
        This method gets the negotiation value based on a given criteria.

        Returns:
            int: value of the negotiation
        """
        _logger.info("Getting the negotiation value for [{}]...".format(self.thread))

        # Since negotiation is a capability of the agent, it is necessary to analyze which skill has been defined. The
        # associated skill interface will be the one from which the value of negotiation can be obtained.
        # Thus, skill is the negotiation criterion for which the ontological instance will be obtained.
        neg_skill_instance = await self.myagent.css_ontology.get_ontology_instance_by_name(self.neg_criteria)

        # The related skill interface will be obtained
        skill_interface = list(neg_skill_instance.get_associated_skill_interface_instances())[0]
        # The AAS element of the skill interface will be used to analyze the skill implementation
        aas_skill_interface_elem = await self.myagent.aas_model.get_object_by_reference(
            skill_interface.get_aas_sme_ref())

        parent_submodel = aas_skill_interface_elem.get_parent_submodel()
        if parent_submodel.check_semantic_id_exist(AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL):
            # In this case, the value need to be obtained through an asset service
            # With the AAS SubmodelElement of the asset interface the related Python class, able to connect to the
            # asset, can be obtained.
            aas_asset_interface_elem = aas_skill_interface_elem.get_associated_asset_interface()
            asset_connection_class = await self.myagent.get_asset_connection_class_by_ref(aas_asset_interface_elem)
            _logger.assetinfo("The Asset connection of the Skill Interface has been obtained.")
            # Now the negotiation value can be obtained through an asset service
            _logger.assetinfo("Obtaining the negotiation value for [{}] through an asset service...".format(self.thread))
            current_neg_value = await asset_connection_class.execute_asset_service(
                interaction_metadata=aas_skill_interface_elem)
            _logger.assetinfo("Negotiation value for [{}] through an asset service obtained: {}.".format(
                self.thread, current_neg_value))
            if not isinstance(current_neg_value, float):
                try:
                    current_neg_value = float(current_neg_value)
                except ValueError as e:
                    # TODO PENSAR OTRAS EXCEPCIONES EN NEGOCIACIONES (durante el asset connection...)
                    _logger.error(e)
                    raise CapabilityRequestExecutionError('Negotiation', "The requested negotiation {} cannot be "
                                                          "executed because the negotiation value returned by the asset does"
                                                          " not have a valid format.".format(self.thread), self)
        else:
            # In this case, the value need to be obtained through an agent service
            try:
                current_neg_value = await self.myagent.agent_services.execute_agent_service_by_id(aas_skill_interface_elem.id_short)
            except (KeyError, ValueError) as e:
                _logger.error(e)
                raise CapabilityRequestExecutionError('Negotiation', "The requested negotiation {} cannot be "
                                                      "executed because the negotiation value cannot be obtained through"
                                                      " the agent service {}.".format(self.thread,
                                                      aas_skill_interface_elem.id_short), self)

        self.neg_value = current_neg_value


    async def send_response_msg_to_sender(self, performative, service_params):
        """
        This method creates and sends a FIPA-ACL message with the given serviceParams and performative.

        Args:
            performative (str): performative according to FIPA-ACL standard.
            service_params (dict): JSON with the serviceParams to be sent in the message.
        """
        acl_msg = inter_aas_interactions_utils.create_inter_smia_response_msg(
            receiver=self.neg_requester_jid,
            thread=self.thread,
            performative=performative,
            ontology=FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION,  # TODO pensar si definir un NegRequest y NegResponse
            service_id='negotiationResult',  # TODO pensar como llamarlo
            service_type='AssetRelatedService',  # TODO ojo si decidimos que es de otro tipo
            service_params=json.dumps(service_params)
        )
        await self.send(acl_msg)


    async def exit_negotiation(self, is_winner):
        """
        This method is executed when the trade has ended, either as a winner or a loser. In any case, all the
        information of the negotiation is added to the global variable with all the information of all the negotiations
         of the agent. The thread is used to differentiate the information of each negotiation, since this is the
         identifier of each one of them.

        Args:
            is_winner (bool): it determines whether the SMIA has been the winner of the negotiation.

        """
        if is_winner:
            _logger.info("The AAS has finished the negotiation with thread [" + self.thread + "] as the winner")
        else:
            _logger.info("The AAS has finished the negotiation with thread [" + self.thread + "] not as the winner")

        # The negotiation information is stored in the global object of the SMIA
        neg_data_json = negotiation_utils.create_neg_json_to_store(neg_requester_jid=self.neg_requester_jid,
                                                                   participants=self.targets,
                                                                   neg_criteria=self.neg_criteria,
                                                                   is_winner=is_winner)
        await self.myagent.save_negotiation_data(thread=self.thread, neg_data=neg_data_json)

        # The information will be stored in the log
        execution_info = {'capName': 'Negotiation', 'capType': CapabilitySkillOntologyUtils.AGENT_CAPABILITY_TYPE,
                          'participants': self.targets, 'criteria': self.neg_criteria, 'winner': str(is_winner)}
        smia_archive_utils.save_completed_svc_log_info(self.requested_timestamp, GeneralUtils.get_current_timestamp(),
                                                       self.svc_req_data, execution_info,
                                                       ServiceTypes.CSS_RELATED_SERVICE)

        # In order to correctly complete the negotiation process, this behavior is removed from the agent.
        self.kill(exit_code=10)
